"""
Restore an algorithm dataset from file, s3 or redis to redis
for ensuring all datasets are ready for Algorithmic backtesting

Supported Datasets:

- ``SA_DATASET_TYPE_ALGO_READY`` - Algorithm-ready datasets
"""

import analysis_engine.load_dataset as load_dataset
import analysis_engine.show_dataset as show_dataset
import analysis_engine.get_data_from_redis_key as redis_utils
import analysis_engine.publish as publish
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import REDIS_ADDRESS
from analysis_engine.consts import DEFAULT_SERIALIZED_DATASETS
from analysis_engine.consts import DATASET_COLLECTION_VERSION
from analysis_engine.consts import SA_DATASET_TYPE_ALGO_READY
from analysis_engine.consts import EMPTY_DF_STR
from analysis_engine.consts import get_percent_done
from spylunking.log.setup_logging import build_colorized_logger

log = build_colorized_logger(
    name=__name__)


def restore_dataset(
        show_summary=True,
        force_restore=False,
        algo_dataset=None,
        dataset_type=SA_DATASET_TYPE_ALGO_READY,
        serialize_datasets=DEFAULT_SERIALIZED_DATASETS,
        path_to_file=None,
        compress=False,
        encoding='utf-8',
        redis_enabled=True,
        redis_key=None,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        redis_serializer='json',
        redis_encoding='utf-8',
        redis_output_db=None,
        s3_enabled=True,
        s3_key=None,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        slack_enabled=False,
        slack_code_block=False,
        slack_full_width=False,
        verbose=False):
    """restore_dataset

    Restore missing dataset nodes in redis from an algorithm-ready
    dataset file on disk. Use this to restore redis from scratch.

    :param show_summary: optional - show a summary of the algorithm-ready
        dataset using ``analysis_engine.show_dataset.show_dataset``
        (default is ``True``)
    :param force_restore: optional - boolean - publish whatever is in
        the algorithm-ready dataset into redis. If ``False`` this will
        ensure that datasets are only set in redis if they are not already
        set
    :param algo_dataset: optional - already loaded algorithm-ready dataset
    :param dataset_type: optional - dataset type
        (default is ``SA_DATASET_TYPE_ALGO_READY``)
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param path_to_file: optional - path to an algorithm-ready dataset
        in a file
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param encoding: optional - string for data encoding

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_key: string - key to save the data in redis
        (default is ``None``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)
    :param redis_serializer: not used yet - support for future
        pickle objects in redis
    :param redis_encoding: format of the encoded key in redis
    :param redis_output_db: optional - integer publish to a separate
        redis database

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_key: string - key to save the data in redis
        (default is ``None``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)

    **(Optional) Slack arguments**

    :param slack_enabled: optional - boolean for
        publishing to slack
    :param slack_code_block: optional - boolean for
        publishing as a code black in slack
    :param slack_full_width: optional - boolean for
        publishing as a to slack using the full
        width allowed

    Additonal arguments

    :param verbose: optional - bool for increasing
        logging
    """

    use_ds = algo_dataset
    redis_host = REDIS_ADDRESS.split(':')[0]
    redis_port = int(REDIS_ADDRESS.split(':')[1])
    if redis_address:
        redis_host = redis_address.split(':')[0]
        redis_port = int(redis_address.split(':')[1])

    if show_summary:
        use_ds = show_dataset.show_dataset(
            dataset_type=dataset_type,
            compress=compress,
            encoding=redis_encoding,
            path_to_file=path_to_file,
            s3_key=s3_key,
            s3_address=s3_address,
            s3_bucket=s3_bucket,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            redis_key=redis_key,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            serialize_datasets=serialize_datasets)

    # end of if show_summary

    if not use_ds:
        log.info(
            'loading from file={} s3={} redis={}'.format(
                path_to_file,
                s3_key,
                redis_key))
        use_ds = load_dataset.load_dataset(
            dataset_type=dataset_type,
            compress=compress,
            encoding=redis_encoding,
            path_to_file=path_to_file,
            s3_key=s3_key,
            s3_address=s3_address,
            s3_bucket=s3_bucket,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            redis_key=redis_key,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            serialize_datasets=serialize_datasets)
    # load if not loaded

    if not use_ds:
        log.error(
            'unable to load a dataset from file={} '
            's3={} redis={}'.format(
                path_to_file,
                s3_key,
                redis_key))
        return None

    log.info('restore - start')
    total_to_restore = 0
    for ticker in use_ds:
        for ds_node in use_ds[ticker]:
            for ds_key in ds_node['data']:
                if ds_key in serialize_datasets:
                    total_to_restore += 1
    # end of counting total_to_restore

    log.info('restore - records={}'.format(total_to_restore))
    num_done = 0
    for ticker in use_ds:
        for ds_node in use_ds[ticker]:
            ds_parent_key = ds_node['id']
            log.info(
                'restore - parent_key={} - {} {}/{}'.format(
                    ds_parent_key,
                    get_percent_done(
                        progress=num_done,
                        total=total_to_restore),
                    num_done,
                    total_to_restore))
            if verbose:
                print(ds_parent_key)

            cache_res = redis_utils.get_data_from_redis_key(
                host=redis_host,
                port=redis_port,
                password=redis_password,
                db=redis_db,
                key=ds_parent_key,
                serializer=redis_serializer,
                encoding=redis_encoding,
                expire=redis_expire,
                label='restore-{}'.format(ds_parent_key))
            should_restore = False
            if (not force_restore and
                    cache_res['status'] == SUCCESS and
                    'data' in cache_res['rec'] and
                    cache_res['rec']['data'] and
                    len(cache_res['rec']['data']) > 10):
                should_restore = False
            else:
                should_restore = True
            if should_restore:
                log.info(
                    ' - parent {} restore'.format(
                        ds_parent_key))
                new_parent_rec = {
                    'exp_date': None,
                    'publish_pricing_update': None,
                    'date': ds_node['date'],
                    'updated': None,
                    'version': DATASET_COLLECTION_VERSION
                }
                for sname in serialize_datasets:
                    if sname in ds_node['data']:
                        if hasattr(
                                ds_node['data'][sname],
                                'index'):
                            new_parent_rec[sname] = \
                                ds_node['data'][sname].to_json(
                                    orient='records',
                                    date_format='iso')
                        else:
                            new_parent_rec[sname] = \
                                ds_node['data'][sname]

                publish.publish(
                    data=new_parent_rec,
                    convert_to_json=False,
                    compress=compress,
                    redis_enabled=True,
                    redis_key=ds_parent_key,
                    redis_db=redis_output_db,
                    redis_address=redis_address,
                    redis_password=redis_password,
                    redis_expire=redis_expire,
                    redis_serializer=redis_serializer,
                    redis_encoding=redis_encoding,
                    s3_enabled=False,
                    output_file=None,
                    verbose=verbose)

            for ds_key in ds_node['data']:
                if ds_key in serialize_datasets:
                    new_key = '{}_{}'.format(
                        ds_parent_key,
                        ds_key)
                    if hasattr(
                            ds_node['data'][ds_key],
                            'index'):
                        loaded_df = ds_node['data'][ds_key]
                        if len(loaded_df.index) > 0:
                            if verbose:
                                print(
                                    ' - checking: {}'.format(
                                        new_key))

                            cache_res = redis_utils.get_data_from_redis_key(
                                host=redis_host,
                                port=redis_port,
                                password=redis_password,
                                db=redis_db,
                                key=new_key,
                                serializer=redis_serializer,
                                encoding=redis_encoding,
                                expire=redis_expire,
                                label='restore-{}'.format(new_key))
                            should_restore = False
                            if (not force_restore and
                                    cache_res['status'] == SUCCESS and
                                    'data' in cache_res['rec'] and
                                    cache_res['rec']['data'] and
                                    len(cache_res['rec']['data']) > 10):
                                should_restore = False
                            else:
                                if (str(cache_res['rec']['data']) !=
                                        EMPTY_DF_STR):
                                    should_restore = True
                            if should_restore:
                                log.info(
                                    'restore nested dataset: {}'.format(
                                        ds_parent_key,
                                        new_key))
                                publish.publish(
                                    data=loaded_df,
                                    is_df=True,
                                    compress=compress,
                                    redis_enabled=True,
                                    redis_key=new_key,
                                    redis_db=redis_output_db,
                                    redis_address=redis_address,
                                    redis_password=redis_password,
                                    redis_expire=redis_expire,
                                    redis_serializer=redis_serializer,
                                    redis_encoding=redis_encoding,
                                    s3_enabled=False,
                                    output_file=None,
                                    verbose=verbose)
                            else:
                                if verbose:
                                    print(
                                        ' - checking: {} - SKIP'.format(
                                            new_key))

                        if verbose:
                            print(' - {} - no data to sync'.format(
                                new_key))
                    # end of is a dataframe
                    # else:
                    # end of handling dataframe vs dictionary

                    num_done += 1
        # end of for all datasets
        print('-----------------------------------')
    # end for all dataset to restore

    log.info(
        'restore - done - num_done={} total={}'.format(
            num_done,
            total_to_restore))

    return use_ds
# end of restore_dataset



"""
Fetch API calls wrapping pyEX

Supported environment variables:

::

    # verbose logging in this module
    export DEBUG_FETCH=1

"""

import pyEX.stocks as pyex_stocks
import analysis_engine.iex.utils as fetch_utils
import analysis_engine.dataset_scrub_utils as scrub_utils
from spylunking.log.setup_logging import build_colorized_logger
from analysis_engine.iex.consts import DATAFEED_DAILY
from analysis_engine.iex.consts import DATAFEED_MINUTE
from analysis_engine.iex.consts import DATAFEED_QUOTE
from analysis_engine.iex.consts import DATAFEED_STATS
from analysis_engine.iex.consts import DATAFEED_PEERS
from analysis_engine.iex.consts import DATAFEED_NEWS
from analysis_engine.iex.consts import DATAFEED_FINANCIALS
from analysis_engine.iex.consts import DATAFEED_EARNINGS
from analysis_engine.iex.consts import DATAFEED_DIVIDENDS
from analysis_engine.iex.consts import DATAFEED_COMPANY

log = build_colorized_logger(
    name=__name__)


def fetch_daily(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_daily

    Fetch the IEX daily data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_DAILY
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - daily - scrub_mode={} args={} '
        'ticker={}'.format(
            label,
            scrub_mode,
            work_dict,
            ticker))

    res = pyex_stocks.chartDF(
        symbol=ticker,
        timeframe=work_dict.get(
            'timeframe',
            '1d'),
        date=work_dict.get(
            'date',
            None))

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_daily


def fetch_minute(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_minute

    Fetch the IEX minute intraday data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_MINUTE
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    from_historical_date = None
    last_close_to_use = None
    dates = []

    if 'from_historical_date' in work_dict:
        from_historical_date = work_dict['from_historical_date']
        last_close_to_use = work_dict['last_close_to_use']
        dates = fetch_utils.get_days_between_dates(
            from_historical_date=work_dict['from_historical_date'],
            last_close_to_use=last_close_to_use)

    log.info(
        '{} - minute - args={} ticker={} '
        'fhdate={} last_close={} dates={}'.format(
            label,
            work_dict,
            ticker,
            from_historical_date,
            last_close_to_use,
            dates))

    res = pyex_stocks.chartDF(
        symbol=ticker,
        timeframe=work_dict.get(
            'timeframe',
            '1m'),
        date=work_dict.get(
            'date',
            None))

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_minute


def fetch_quote(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_quote

    Fetch the IEX quote data for a ticker and
    return as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_QUOTE
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - quote - args={} ticker={}'.format(
            label,
            work_dict,
            ticker))

    res = pyex_stocks.quoteDF(
        symbol=ticker)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_quote


def fetch_stats(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_stats

    Fetch the IEX statistics data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_STATS
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - stats - args={} ticker={}'.format(
            label,
            work_dict,
            ticker))

    res = pyex_stocks.stockStatsDF(
        symbol=ticker)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_stats


def fetch_peers(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_peers

    Fetch the IEX peers data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_PEERS
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - peers - args={} ticker={}'.format(
            label,
            work_dict,
            ticker))

    res = pyex_stocks.peersDF(
        symbol=ticker)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_peers


def fetch_news(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_news

    Fetch the IEX news data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_NEWS
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - news - args={} ticker={}'.format(
            label,
            work_dict,
            ticker))

    res = pyex_stocks.newsDF(
        symbol=ticker,
        count=50)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_news


def fetch_financials(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_financials

    Fetch the IEX financial data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_FINANCIALS
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - financials - args={} ticker={}'.format(
            label,
            work_dict,
            ticker))

    res = pyex_stocks.financialsDF(
        symbol=ticker)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_financials


def fetch_earnings(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_earnings

    Fetch the IEX earnings data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_EARNINGS
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - earnings - args={} ticker={}'.format(
            label,
            work_dict,
            ticker))

    res = pyex_stocks.earningsDF(
        symbol=ticker)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_earnings


def fetch_dividends(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_dividends

    Fetch the IEX dividends data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_DIVIDENDS
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    timeframe = work_dict.get(
        'timeframe',
        '2y')
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - dividends - args={} ticker={} '
        'timeframe={}'.format(
            label,
            work_dict,
            ticker,
            timeframe))

    res = pyex_stocks.dividendsDF(
        symbol=ticker,
        timeframe=timeframe)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_dividends


def fetch_company(
        work_dict,
        scrub_mode='NO_SORT'):
    """fetch_company

    Fetch the IEX company data for a ticker and
    return it as a ``pandas.DataFrame``.

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    datafeed_type = DATAFEED_COMPANY
    ticker = work_dict.get(
        'ticker',
        None)
    label = work_dict.get(
        'label',
        None)
    use_date = work_dict.get(
        'use_date',
        None)

    log.info(
        '{} - company - args={} ticker={}'.format(
            label,
            label,
            work_dict,
            ticker))

    res = pyex_stocks.companyDF(
        symbol=ticker)

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=use_date,
        df=res)

    return scrubbed_df
# end of fetch_company



"""
Algo data helper for mapping indicator category
to an integer label value for downstream dataset predictions
"""
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_category_as_int(
        node,
        label=None):
    """get_category_as_int

    Helper for converting feature labels to numeric values

    :param node: convert the dictionary's ``category``
        string to the integer mapped value
    """
    if not label:
        label = 'get_category_as_int'
    ind_category_str = node.get(
        'category',
        'momentum')
    if not ind_category_str:
        return ae_consts.INDICATOR_CATEGORY_UNKNOWN
    elif ind_category_str == 'momentum':
        return ae_consts.INDICATOR_CATEGORY_MOMENTUM
    elif ind_category_str == 'overlap':
        return ae_consts.INDICATOR_CATEGORY_OVERLAP
    elif ind_category_str == 'price':
        return ae_consts.INDICATOR_CATEGORY_PRICE
    elif ind_category_str == 'volume':
        return ae_consts.INDICATOR_CATEGORY_VOLUME
    elif ind_category_str == 'volatility':
        return ae_consts.INDICATOR_CATEGORY_VOLATILITY
    elif ind_category_str == 'single_call':
        return ae_consts.INDICATOR_CATEGORY_SINGLE_CALL
    elif ind_category_str == 'single_put':
        return ae_consts.INDICATOR_CATEGORY_SINGLE_PUT
    elif ind_category_str == 'bull_call':
        return ae_consts.INDICATOR_CATEGORY_BULL_CALL
    elif ind_category_str == 'bear_put':
        return ae_consts.INDICATOR_CATEGORY_BEAR_PUT
    elif ind_category_str == 'quarterly':
        return ae_consts.INDICATOR_CATEGORY_QUARTERLY
    elif ind_category_str == 'yearly':
        return ae_consts.INDICATOR_CATEGORY_YEARLY
    elif ind_category_str == 'income_statement':
        return ae_consts.INDICATOR_CATEGORY_INCOME_STMT
    elif ind_category_str == 'cash_flow':
        return ae_consts.INDICATOR_CATEGORY_CASH_FLOW
    elif ind_category_str == 'balance_sheet':
        return ae_consts.INDICATOR_CATEGORY_BALANCE_SHEET
    elif ind_category_str == 'press_release':
        return ae_consts.INDICATOR_CATEGORY_PRESS_RELEASE
    elif ind_category_str == 'news':
        return ae_consts.INDICATOR_CATEGORY_NEWS
    elif ind_category_str == 'earnings':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'splits':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'reverse_splits':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'distributions':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'spinoffs':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'merger_acq':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'exchange_inclusion':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'exchange_exclusion':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'clinical_trial_positive':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'clinical_trial_negative':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'short_sellers':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'custom':
        return ae_consts.INDICATOR_CATEGORY_CUSTOM
    elif ind_category_str == 'news':
        return ae_consts.INDICATOR_CATEGORY_NEWS
    elif ind_category_str == 'earnings':
        return ae_consts.INDICATOR_CATEGORY_EARNINGS
    elif ind_category_str == 'csuite':
        return ae_consts.INDICATOR_CATEGORY_CSUITE
    elif ind_category_str == 'splits':
        return ae_consts.INDICATOR_CATEGORY_SPLITS
    elif ind_category_str == 'reverse_splits':
        return ae_consts.INDICATOR_CATEGORY_REVERSE_SPLITS
    elif ind_category_str == 'distributions':
        return ae_consts.INDICATOR_CATEGORY_DISTRIBUTIONS
    elif ind_category_str == 'spinoffs':
        return ae_consts.INDICATOR_CATEGORY_SPINOFFS
    elif ind_category_str == 'merger_acq':
        return ae_consts.INDICATOR_CATEGORY_MERGER_ACQ
    elif ind_category_str == 'exchange_inclusion':
        return ae_consts.INDICATOR_CATEGORY_EXCHANGE_INCLUSION
    elif ind_category_str == 'exchange_exclusion':
        return ae_consts.INDICATOR_CATEGORY_EXCHANGE_EXCLUSION
    elif ind_category_str == 'trial_positive':
        return ae_consts.INDICATOR_CATEGORY_TRIAL_POSITIVE
    elif ind_category_str == 'trial_negative':
        return ae_consts.INDICATOR_CATEGORY_TRIAL_NEGATIVE
    elif ind_category_str == 'short_sellers':
        return ae_consts.INDICATOR_CATEGORY_SHORT_SELLERS
    elif ind_category_str == 'real_estate':
        return ae_consts.INDICATOR_CATEGORY_REAL_ESTATE
    elif ind_category_str == 'housing':
        return ae_consts.INDICATOR_CATEGORY_HOUSING
    elif ind_category_str == 'pipeline':
        return ae_consts.INDICATOR_CATEGORY_PIPELINE
    elif ind_category_str == 'construction':
        return ae_consts.INDICATOR_CATEGORY_CONSTRUCTION
    elif ind_category_str == 'fed':
        return ae_consts.INDICATOR_CATEGORY_FED
    else:
        return ae_consts.INDICATOR_CATEGORY_UNKNOWN
        if label:
            log.error(
                'unsupported indicator '
                'uses_dataset={} defaulting to "unknown"'.format(
                    ind_category_str))
        else:
            log.error(
                '{} - unsupported indicator '
                'uses_dataset={} defaulting to "unknown"'.format(
                    label,
                    ind_category_str))
    # end of supported indicator dataset types
# end of get_category_as_int



"""
Helper for loading datasets from a file

**Supported environment variables**

::

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json
"""

import analysis_engine.prepare_dict_for_algo as prepare_utils
from analysis_engine.consts import DEFAULT_SERIALIZED_DATASETS
from spylunking.log.setup_logging import build_colorized_logger

log = build_colorized_logger(
    name=__name__)


def load_algo_dataset_from_file(
        path_to_file,
        serialize_datasets=DEFAULT_SERIALIZED_DATASETS,
        compress=False,
        encoding='utf-8'):
    """load_algo_dataset_from_file

    Load an algorithm-ready dataset for algorithm backtesting
    from a local file

    :param path_to_file: string - path to file holding an
        algorithm-ready dataset
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param encoding: optional - string for data encoding
    """
    log.debug('start')
    data_from_file = None
    with open(path_to_file, 'r') as cur_file:
        data_from_file = cur_file.read()

    if not data_from_file:
        log.error('missing data from file={}'.format(
            path_to_file))
        return None

    return prepare_utils.prepare_dict_for_algo(
        data=data_from_file,
        compress=compress,
        convert_to_dict=True,
        encoding=encoding)
# end of load_algo_dataset_from_file



"""
Get defaults fields for supported teyps of data
"""
from analysis_engine.iex.consts import DATAFEED_DAILY
from analysis_engine.iex.consts import DATAFEED_QUOTE
from analysis_engine.iex.consts import DATAFEED_STATS
from analysis_engine.iex.consts import DATAFEED_PEERS
from analysis_engine.iex.consts import DATAFEED_NEWS
from analysis_engine.iex.consts import DATAFEED_FINANCIALS
from analysis_engine.iex.consts import DATAFEED_EARNINGS
from analysis_engine.iex.consts import DATAFEED_DIVIDENDS
from analysis_engine.iex.consts import DATAFEED_COMPANY
from spylunking.log.setup_logging import build_colorized_logger

log = build_colorized_logger(
    name=__name__)


def get_default_fields(
            field):
    """get_default_fields

    :param field: types of data to get
    """
    use_field = str(field).lower()
    if use_field == 'daily' or field == DATAFEED_DAILY:
        return ['KEY', 'date']
    elif use_field == 'quote' or field == DATAFEED_QUOTE:
        return ['KEY', 'date', 'minute']
    elif use_field == 'stats' or field == DATAFEED_STATS:
        return ['KEY']
    elif use_field == 'peers' or field == DATAFEED_PEERS:
        return ['KEY', 'peer']
    elif use_field == 'news' or field == DATAFEED_NEWS:
        return ['KEY', 'datetime']
    elif use_field == 'financials' or field == DATAFEED_FINANCIALS:
        return ['KEY', 'reportDate']
    elif use_field == 'earnings' or field == DATAFEED_EARNINGS:
        return ['KEY', 'EPSReportDate']
    elif use_field == 'dividends' or field == DATAFEED_DIVIDENDS:
        return ['KEY', 'exDate']
    elif use_field == 'company' or field == DATAFEED_COMPANY:
        return ['KEY']
    else:
        log.error(
            'get_default_fields({}) is not a supported '
            'field'.format(
                field))
        raise NotImplemented
# end of get_default_fields



"""
TA-Lib wrappers
"""

# for unittests, allow passing the mocks into the runtime if not found
try:
    import talib
except Exception:
    import analysis_engine.mocks.mock_talib as talib
# end of loading talib or mocks
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def WILLR(
        high=None,
        low=None,
        close=None,
        timeperiod=None,
        verbose=False):
    """WILLR

    Wrapper for talib.WILLR for running unittests
    on ci/cd tools that do not provide talib

    :param high: hostname
    :param low: port
    :param close: password
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'willr - start')
    return talib.WILLR(
        high,
        low,
        close,
        timeperiod)
# end of WILLR



"""
Build a result dictionary
"""

from analysis_engine.consts import NOT_RUN


def build_result(
        status=NOT_RUN,
        err=None,
        rec=None):
    """build_result

    Common result builder for helping standardize
    results per function and unittesting

    :param status: starting status value from ``consts.py``
    :param err: optional error message
    :param rec: dictionary for the result
    """

    res = {
        'status': status,
        'err': err,
        'rec': rec
    }
    if not res:
        res['rec'] = {}

    return res
# end of build_result



try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

import requests
import tornado
import ujson
import pandas as pd
import pyEX as p
import string
import warnings
warnings.filterwarnings("ignore")  # noqa
from spylunking.log.setup_logging import build_colorized_logger
from trading_calendars import get_calendar
from datetime import datetime
from datetime import date
from datetime import timedelta
from dateutil.relativedelta import relativedelta


_TRANSLATOR = str.maketrans('', '', string.punctuation)
_OVERRIDES = {
    'PCLN': 'BKNG'
}

log = build_colorized_logger(
    name=__name__)


def parse_args(argv):
    args = []
    kwargs = {}
    for arg in argv:
        if '--' not in arg and '-' not in arg:
            log.debug('ignoring argument: %s', arg)
            continue
        if '=' in arg:
            k, v = arg.replace('-', '').split('=')
            kwargs[k] = v
        else:
            args.append(arg.replace('-', ''))
    return args, kwargs


def parse_body(req, **fields):
    try:
        data = tornado.escape.json_decode(req.body)
    except ValueError:
        data = {}
    return data


def safe_get(path, *args, **kwargs):
    try:
        log.debug('GET: %s' % path)
        resp = requests.get(path, *args, **kwargs).text
        # log.debug('GET_RESPONSE: %s' % resp)
        return ujson.loads(resp)
    except ConnectionRefusedError:
        return {}


def safe_post(path, *args, **kwargs):
    try:
        log.debug('POST: %s' % path)
        resp = requests.post(path, *args, **kwargs).text
        # log.debug('POST_RESPONSE: %s' % resp)
        return ujson.loads(resp)
    except ConnectionRefusedError:
        return {}


def safe_post_cookies(path, *args, **kwargs):
    try:
        log.debug('POST: %s' % path)
        resp = requests.post(path, *args, **kwargs)
        # log.debug('POST_RESPONSE: %s' % resp.text)
        return ujson.loads(resp.text), resp.cookies
    except ConnectionRefusedError:
        return {}, None


def construct_path(host, method):
    return urljoin(host, method)


def symbols():
    return p.symbolsDF().index.values.tolist()


def symbols_map():
    ret = {}
    for x in symbols():
        ret[x] = x
        new_x = x.translate(_TRANSLATOR)
        if new_x not in ret:
            ret[new_x] = x
    for k, v in _OVERRIDES.items():
        ret[k] = v
    return ret


def today():
    """today starts at 4pm the previous close"""
    today = date.today()
    return datetime(
        year=today.year,
        month=today.month,
        day=today.day)


def this_week():
    """start of week"""
    return today() - timedelta(days=datetime.today().isoweekday() % 7)


def last_close():
    """last close"""
    today = date.today()
    close = datetime(
        year=today.year,
        month=today.month,
        day=today.day, hour=16)

    if today.weekday() == 5:
        return close - timedelta(days=1)
    elif today.weekday() == 6:
        return close - timedelta(days=2)
    else:
        if datetime.now().hour < 16:
            close -= timedelta(days=1)
            if close.weekday() == 5:  # saturday
                return close - timedelta(days=1)
            elif close.weekday() == 6:  # sunday
                return close - timedelta(days=2)
            return close
        return close


def yesterday():
    """yesterday is anytime before the previous 4pm close"""
    today = date.today()

    if today.weekday() == 0:  # monday
        return datetime(
            year=today.year,
            month=today.month,
            day=today.day) - timedelta(days=3)
    elif today.weekday() == 6:  # sunday
        return datetime(
            year=today.year,
            month=today.month,
            day=today.day) - timedelta(days=2)
    return datetime(
        year=today.year,
        month=today.month,
        day=today.day) - timedelta(days=1)


def last_month():
    """last_month is one month before today"""
    today = date.today()
    last_month = datetime(
        year=today.year,
        month=today.month,
        day=today.day) - relativedelta(months=1)

    if last_month.weekday() == 5:
        last_month -= timedelta(days=1)
    elif last_month.weekday() == 6:
        last_month -= timedelta(days=2)
    return last_month


def six_months():
    """six_months is six months before today"""
    today = date.today()
    six_months = datetime(
        year=today.year,
        month=today.month,
        day=today.day) - relativedelta(months=6)

    if six_months.weekday() == 5:
        six_months -= timedelta(days=1)
    elif six_months.weekday() == 6:
        six_months -= timedelta(days=2)
    return six_months


def three_months():
    """three_months"""
    today = date.today()
    six_months = datetime(
        year=today.year,
        month=today.month,
        day=today.day) - relativedelta(months=3)

    if six_months.weekday() == 5:
        six_months -= timedelta(days=1)
    elif six_months.weekday() == 6:
        six_months -= timedelta(days=2)
    return six_months


def never():
    """never"""
    return datetime(
        year=1,
        month=1,
        day=1)


def append(df1, df2):
    """append

    :param df1:
    :param df2:
    """
    merged = pd.concat([df1, df2])
    return merged[~merged.index.duplicated(keep='first')]


def holidays():
    """holidays"""
    return get_calendar(
        'NYSE').regular_holidays.holidays().to_pydatetime().tolist()


def business_days(start, end=last_close()):
    """business_days"""
    ret = []
    while start < last_close():
        is_a_business_day = (
            start not in holidays() and
            start.weekday() != 6 and start.weekday() != 5)
        if is_a_business_day:
            ret.append(start)
        start += timedelta(days=1)
    return ret


def get_days_between_dates(
        from_historical_date,
        last_close_to_use=None):
    """get_days_between_dates

    :param from_historical_date: historical date in time to start walking
                                 forward until the last close datetime
    :param last_close_to_use: starting date in time (left leg of window)
    """
    use_last_close = last_close_to_use
    if not use_last_close:
        use_last_close = last_close()

    dates = []
    while from_historical_date < last_close_to_use:
        dates.append(from_historical_date)
        from_historical_date += datetime.timedelta(
            days=1)
    return dates
# end of get_days_between_dates



"""
Parse data from IEX

Supported environment variables:

::

    export DEBUG_IEX_DATA=1

"""

import datetime
import copy
import analysis_engine.build_result as build_result
import analysis_engine.api_requests as api_requests
import analysis_engine.iex.fetch_data as iex_fetch_data
import analysis_engine.work_tasks.publish_pricing_update as \
    publisher
from spylunking.log.setup_logging import build_colorized_logger
from analysis_engine.consts import TICKER
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import NOT_RUN
from analysis_engine.consts import ERR
from analysis_engine.consts import NOT_SET
from analysis_engine.consts import get_status
from analysis_engine.consts import ev
from analysis_engine.iex.consts import FETCH_DAILY
from analysis_engine.iex.consts import FETCH_MINUTE
from analysis_engine.iex.consts import FETCH_QUOTE
from analysis_engine.iex.consts import FETCH_STATS
from analysis_engine.iex.consts import FETCH_PEERS
from analysis_engine.iex.consts import FETCH_NEWS
from analysis_engine.iex.consts import FETCH_FINANCIALS
from analysis_engine.iex.consts import FETCH_EARNINGS
from analysis_engine.iex.consts import FETCH_DIVIDENDS
from analysis_engine.iex.consts import FETCH_COMPANY

log = build_colorized_logger(
    name=__name__)


def get_data_from_iex(
        work_dict):
    """get_data_from_iex

    Get pricing from iex

    :param work_dict: request dictionary
    """
    label = 'get_data_from_iex'

    log.info(
        'task - {} - start '
        'work_dict={}'.format(
            label,
            work_dict))

    rec = {
        'data': None,
        'updated': None
    }
    res = {
        'status': NOT_RUN,
        'err': None,
        'rec': rec
    }

    ticker = None
    field = None
    ft_type = None

    try:

        ticker = work_dict.get(
            'ticker',
            TICKER)
        field = work_dict.get(
            'field',
            'daily')
        ft_type = work_dict.get(
            'ft_type',
            None)
        ft_str = str(ft_type).lower()
        label = work_dict.get(
            'label',
            label)
        orient = work_dict.get(
            'orient',
            'records')

        iex_req = None
        if ft_type == FETCH_DAILY or ft_str == 'daily':
            ft_type == FETCH_DAILY
            iex_req = api_requests.build_iex_fetch_daily_request(
                label=label)
        elif ft_type == FETCH_MINUTE or ft_str == 'minute':
            ft_type == FETCH_MINUTE
            iex_req = api_requests.build_iex_fetch_minute_request(
                label=label)
        elif ft_type == FETCH_QUOTE or ft_str == 'quote':
            ft_type == FETCH_QUOTE
            iex_req = api_requests.build_iex_fetch_quote_request(
                label=label)
        elif ft_type == FETCH_STATS or ft_str == 'stats':
            ft_type == FETCH_STATS
            iex_req = api_requests.build_iex_fetch_stats_request(
                label=label)
        elif ft_type == FETCH_PEERS or ft_str == 'peers':
            ft_type == FETCH_PEERS
            iex_req = api_requests.build_iex_fetch_peers_request(
                label=label)
        elif ft_type == FETCH_NEWS or ft_str == 'news':
            ft_type == FETCH_NEWS
            iex_req = api_requests.build_iex_fetch_news_request(
                label=label)
        elif ft_type == FETCH_FINANCIALS or ft_str == 'financials':
            ft_type == FETCH_FINANCIALS
            iex_req = api_requests.build_iex_fetch_financials_request(
                label=label)
        elif ft_type == FETCH_EARNINGS or ft_str == 'earnings':
            ft_type == FETCH_EARNINGS
            iex_req = api_requests.build_iex_fetch_earnings_request(
                label=label)
        elif ft_type == FETCH_DIVIDENDS or ft_str == 'dividends':
            ft_type == FETCH_DIVIDENDS
            iex_req = api_requests.build_iex_fetch_dividends_request(
                label=label)
        elif ft_type == FETCH_COMPANY or ft_str == 'company':
            ft_type == FETCH_COMPANY
            iex_req = api_requests.build_iex_fetch_company_request(
                label=label)
        else:
            log.error(
                '{} - unsupported ft_type={} ft_str={} ticker={}'.format(
                    label,
                    ft_type,
                    ft_str,
                    ticker))
            raise NotImplemented
        # if supported fetch request type

        clone_keys = [
            'ticker',
            's3_address',
            's3_bucket',
            's3_key',
            'redis_address',
            'redis_db',
            'redis_password',
            'redis_key'
        ]

        for k in clone_keys:
            iex_req[k] = work_dict.get(
                k,
                '{}-missing-in-{}'.format(
                    k,
                    label))
        # end of cloning keys

        if not iex_req:
            err = (
                '{} - ticker={} did not build an IEX request '
                'for work={}'.format(
                    label,
                    iex_req['ticker'],
                    work_dict))
            log.error(err)
            res = build_result.build_result(
                status=ERR,
                err=err,
                rec=rec)
            return res
        else:
            log.info(
                '{} - ticker={} field={} '
                'orient={} fetch'.format(
                    label,
                    iex_req['ticker'],
                    field,
                    orient))
        # if invalid iex request

        df = None
        try:
            if 'from' in work_dict:
                iex_req['from'] = datetime.datetime.strptime(
                    '%Y-%m-%d %H:%M:%S',
                    work_dict['from'])
            df = iex_fetch_data.fetch_data(
                work_dict=iex_req,
                fetch_type=ft_type)
            rec['data'] = df.to_json(
                orient=orient,
                date_format='iso')
            rec['updated'] = datetime.datetime.utcnow().strftime(
                '%Y-%m-%d %H:%M:%S')
        except Exception as f:
            log.error(
                '{} - ticker={} field={} failed fetch_data '
                'with ex={}'.format(
                    label,
                    iex_req['ticker'],
                    ft_type,
                    f))
        # end of try/ex

        if ev('DEBUG_IEX_DATA', '0') == '1':
            log.info(
                '{} ticker={} field={} data={} to_json'.format(
                    label,
                    iex_req['ticker'],
                    field,
                    rec['data']))
        else:
            log.info(
                '{} ticker={} field={} to_json'.format(
                    label,
                    iex_req['ticker'],
                    field))
        # end of if/else found data

        upload_and_cache_req = copy.deepcopy(iex_req)
        upload_and_cache_req['celery_disabled'] = True
        upload_and_cache_req['data'] = rec['data']
        if not upload_and_cache_req['data']:
            upload_and_cache_req['data'] = '{}'
        use_field = field
        if use_field == 'news':
            use_field = 'news1'
        if 'redis_key' in work_dict:
            upload_and_cache_req['redis_key'] = '{}_{}'.format(
                work_dict.get(
                    'redis_key',
                    iex_req['redis_key']),
                use_field)
        if 's3_key' in work_dict:
            upload_and_cache_req['s3_key'] = '{}_{}'.format(
                work_dict.get(
                    's3_key',
                    iex_req['s3_key']),
                use_field)

        try:
            update_res = publisher.run_publish_pricing_update(
                work_dict=upload_and_cache_req)
            update_status = update_res.get(
                'status',
                NOT_SET)
            log.info(
                '{} publish update status={} data={}'.format(
                    label,
                    get_status(status=update_status),
                    update_res))
        except Exception as f:
            err = (
                '{} - failed to upload iex data={} to '
                'to s3_key={} and redis_key={}'.format(
                    label,
                    upload_and_cache_req,
                    upload_and_cache_req['s3_key'],
                    upload_and_cache_req['redis_key']))
            log.error(err)
        # end of try/ex to upload and cache

        if not rec['data']:
            log.info(
                '{} - ticker={} no IEX data field={} to publish'.format(
                    label,
                    iex_req['ticker'],
                    field))
        # end of if/else

        res = build_result.build_result(
            status=SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        res = build_result.build_result(
            status=ERR,
            err=(
                'failed - get_data_from_iex '
                'dict={} with ex={}').format(
                    work_dict,
                    e),
            rec=rec)
    # end of try/ex

    log.info(
        'task - get_data_from_iex done - '
        '{} - status={} err={}'.format(
            label,
            get_status(res['status']),
            res['err']))

    return res
# end of get_data_from_iex



"""
Utility to output an n-level nested dictionary as a CSV
"""

import csv
import os
from spylunking.log.setup_logging import build_colorized_logger


log = build_colorized_logger(
    name=__name__)


def flatten_dict(
        data,
        parent_key='',
        sep='_'):
    """flatten_dict

    Flatten an n-level nested dictionary for csv output

    :param data: Dictionary to be parsed
    :param parent_key: The nested parent key
    :param sep: The separator to use between keys
    """
    items = []
    for key, value in data.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, dict):
            items.extend(flatten_dict(value, new_key, sep=sep).items())
        elif isinstance(value, list):
            for idx, val in enumerate(value):
                temp_key = '{}_{}'.format(new_key, idx)
                items.extend(flatten_dict(
                    val,
                    temp_key,
                    sep=sep).items())
        else:
            items.append((new_key, value))
    return dict(items)
# end of flatten_dict


def dict_to_csv(
        data,
        filename='test'):
    """dict_to_csv

    Convert a dictionary to an output CSV

    :param data: Dictionary to be converted
    :param filename: The name of the CSV to produce
    """
    noext_filename = os.path.splitext(filename)[0]
    log.info(('START dict={} conversion to csv={}.csv').format(
        data,
        noext_filename))
    flattened_data = flatten_dict(data)
    with open('{}.csv'.format(noext_filename), 'w') as f:
        w = csv.DictWriter(f, flattened_data.keys())
        w.writeheader()
        w.writerow(flattened_data)
# end of dict_to_csv



#!/usr/bin/env python

"""

Publish the contents of an S3 key to a
Redis key

Steps:
------

1) Parse arguments
2) Download pricing data as a Celery task
3) Publish pricing data as a Celery tasks
4) Coming Soon - Start buy/sell analysis as Celery task(s)

"""

import argparse
import analysis_engine.work_tasks.publish_from_s3_to_redis \
    as task_publisher
from celery import signals
from spylunking.log.setup_logging import build_colorized_logger
from analysis_engine.work_tasks.get_celery_app import get_celery_app
from analysis_engine.api_requests import build_publish_from_s3_to_redis_request
from analysis_engine.consts import LOG_CONFIG_PATH
from analysis_engine.consts import TICKER
from analysis_engine.consts import TICKER_ID
from analysis_engine.consts import WORKER_BROKER_URL
from analysis_engine.consts import WORKER_BACKEND_URL
from analysis_engine.consts import WORKER_CELERY_CONFIG_MODULE
from analysis_engine.consts import INCLUDE_TASKS
from analysis_engine.consts import SSL_OPTIONS
from analysis_engine.consts import TRANSPORT_OPTIONS
from analysis_engine.consts import S3_ACCESS_KEY
from analysis_engine.consts import S3_SECRET_KEY
from analysis_engine.consts import S3_REGION_NAME
from analysis_engine.consts import S3_ADDRESS
from analysis_engine.consts import S3_SECURE
from analysis_engine.consts import S3_BUCKET
from analysis_engine.consts import S3_KEY
from analysis_engine.consts import REDIS_ADDRESS
from analysis_engine.consts import REDIS_KEY
from analysis_engine.consts import REDIS_PASSWORD
from analysis_engine.consts import REDIS_DB
from analysis_engine.consts import REDIS_EXPIRE
from analysis_engine.consts import get_status
from analysis_engine.consts import ppj
from analysis_engine.consts import is_celery_disabled


# Disable celery log hijacking
# https://github.com/celery/celery/issues/2509
@signals.setup_logging.connect
def setup_celery_logging(**kwargs):
    pass


log = build_colorized_logger(
    name='pub-from-s3-to-redis',
    log_config_path=LOG_CONFIG_PATH)


def publish_from_s3_to_redis():
    """publish_from_s3_to_redis

    Download an S3 key and publish it's contents to Redis

    """

    log.info(
        'start - publish_from_s3_to_redis')

    parser = argparse.ArgumentParser(
        description=(
            'Download and store the latest stock pricing, '
            'news, and options chain data '
            'and store it in S3 and Redis. '
            'Once stored, this will also '
            'start the buy and sell trading analysis.'))
    parser.add_argument(
        '-t',
        help=(
            'ticker'),
        required=True,
        dest='ticker')
    parser.add_argument(
        '-i',
        help=(
            'optional - ticker id '
            'not used without a database'),
        required=False,
        dest='ticker_id')
    parser.add_argument(
        '-l',
        help=(
            'optional - path to the log config file'),
        required=False,
        dest='log_config_path')
    parser.add_argument(
        '-b',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='broker_url')
    parser.add_argument(
        '-B',
        help=(
            'optional - backend url for Celery'),
        required=False,
        dest='backend_url')
    parser.add_argument(
        '-k',
        help=(
            'optional - s3 access key'),
        required=False,
        dest='s3_access_key')
    parser.add_argument(
        '-s',
        help=(
            'optional - s3 secret key'),
        required=False,
        dest='s3_secret_key')
    parser.add_argument(
        '-a',
        help=(
            'optional - s3 address format: <host:port>'),
        required=False,
        dest='s3_address')
    parser.add_argument(
        '-S',
        help=(
            'optional - s3 ssl or not'),
        required=False,
        dest='s3_secure')
    parser.add_argument(
        '-u',
        help=(
            'optional - s3 bucket name'),
        required=False,
        dest='s3_bucket_name')
    parser.add_argument(
        '-g',
        help=(
            'optional - s3 region name'),
        required=False,
        dest='s3_region_name')
    parser.add_argument(
        '-p',
        help=(
            'optional - redis_password'),
        required=False,
        dest='redis_password')
    parser.add_argument(
        '-r',
        help=(
            'optional - redis_address format: <host:port>'),
        required=False,
        dest='redis_address')
    parser.add_argument(
        '-n',
        help=(
            'optional - redis and s3 key name'),
        required=False,
        dest='keyname')
    parser.add_argument(
        '-m',
        help=(
            'optional - redis database number (0 by default)'),
        required=False,
        dest='redis_db')
    parser.add_argument(
        '-x',
        help=(
            'optional - redis expiration in seconds'),
        required=False,
        dest='redis_expire')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    ticker = TICKER
    ticker_id = TICKER_ID
    ssl_options = SSL_OPTIONS
    transport_options = TRANSPORT_OPTIONS
    broker_url = WORKER_BROKER_URL
    backend_url = WORKER_BACKEND_URL
    celery_config_module = WORKER_CELERY_CONFIG_MODULE
    include_tasks = INCLUDE_TASKS
    s3_access_key = S3_ACCESS_KEY
    s3_secret_key = S3_SECRET_KEY
    s3_region_name = S3_REGION_NAME
    s3_address = S3_ADDRESS
    s3_secure = S3_SECURE
    s3_bucket_name = S3_BUCKET
    s3_key = S3_KEY
    redis_address = REDIS_ADDRESS
    redis_key = REDIS_KEY
    redis_password = REDIS_PASSWORD
    redis_db = REDIS_DB
    redis_expire = REDIS_EXPIRE
    debug = False

    if args.ticker:
        ticker = args.ticker.upper()
    if args.ticker_id:
        ticker = args.ticker_id
    if args.broker_url:
        broker_url = args.broker_url
    if args.backend_url:
        backend_url = args.backend_url
    if args.s3_access_key:
        s3_access_key = args.s3_access_key
    if args.s3_secret_key:
        s3_secret_key = args.s3_secret_key
    if args.s3_region_name:
        s3_region_name = args.s3_region_name
    if args.s3_address:
        s3_address = args.s3_address
    if args.s3_secure:
        s3_secure = args.s3_secure
    if args.s3_bucket_name:
        s3_bucket_name = args.s3_bucket_name
    if args.keyname:
        s3_key = args.keyname
        redis_key = args.keyname
    if args.redis_address:
        redis_address = args.redis_address
    if args.redis_password:
        redis_password = args.redis_password
    if args.redis_db:
        redis_db = args.redis_db
    if args.redis_expire:
        redis_expire = args.redis_expire
    if args.debug:
        debug = True

    work = build_publish_from_s3_to_redis_request()

    work['ticker'] = ticker
    work['ticker_id'] = ticker_id
    work['s3_bucket'] = s3_bucket_name
    work['s3_key'] = s3_key
    work['redis_key'] = redis_key
    work['s3_access_key'] = s3_access_key
    work['s3_secret_key'] = s3_secret_key
    work['s3_region_name'] = s3_region_name
    work['s3_address'] = s3_address
    work['s3_secure'] = s3_secure
    work['redis_address'] = redis_address
    work['redis_password'] = redis_password
    work['redis_db'] = redis_db
    work['redis_expire'] = redis_expire
    work['debug'] = debug
    work['label'] = 'ticker={}'.format(
        ticker)

    path_to_tasks = 'analysis_engine.work_tasks'
    task_name = (
        '{}.publish_from_s3_to_redis.publish_from_s3_to_redis'.format(
            path_to_tasks))
    task_res = None
    if is_celery_disabled():
        work['celery_disabled'] = True
        log.debug(
            'starting without celery work={}'.format(
                ppj(work)))
        task_res = task_publisher.publish_from_s3_to_redis(
            work_dict=work)
        if debug:
            log.info(
                'done - result={} '
                'task={} status={} '
                'err={} label={}'.format(
                    ppj(task_res),
                    task_name,
                    get_status(status=task_res['status']),
                    task_res['err'],
                    work['label']))
        else:
            log.info(
                'done - result '
                'task={} status={} '
                'err={} label={}'.format(
                    task_name,
                    get_status(status=task_res['status']),
                    task_res['err'],
                    work['label']))
        # if/else debug
    else:
        log.info(
            'connecting to broker={} backend={}'.format(
                broker_url,
                backend_url))

        # Get the Celery app
        app = get_celery_app(
            name=__name__,
            auth_url=broker_url,
            backend_url=backend_url,
            path_to_config_module=celery_config_module,
            ssl_options=ssl_options,
            transport_options=transport_options,
            include_tasks=include_tasks)

        log.info(
            'calling task={} - work={}'.format(
                task_name,
                ppj(work)))
        job_id = app.send_task(
            task_name,
            (work,))
        log.info(
            'calling task={} - success job_id={}'.format(
                task_name,
                job_id))
    # end of if/else
# end of publish_from_s3_to_redis


if __name__ == '__main__':
    publish_from_s3_to_redis()



"""
**Handle Pricing Update Task**

Get the latest stock news, quotes and options chains for a
ticker and publish the values to redis and S3 for downstream analysis.

Writes pricing updates to S3 and Redis by
building a list of publishing sub-task:

**Sample work_dict request for this method**

`analysis_engine.api_requests.publish_pricing_update <https://
github.com/AlgoTraders/stock-analysis-engine/blob/master/ana
lysis_engine/api_requests.py#L218>`__

::

    work = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'prepared_s3_key': s3_prepared_key,
        'prepared_s3_bucket': s3_prepared_bucket_name,
        'prepared_redis_key': redis_prepared_key,
        'ignore_columns': ignore_columns,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

.. tip:: This task uses the `analysis_engine.work_tasks.
    custom_task.CustomTask class <https://github.com/A
    lgoTraders/stock-analysis-engine/blob/master/anal
    ysis_engine/work_tasks/custom_task.py>`__ for
    task event handling.

**Supported Environment Variables**

::

    export DEBUG_RESULTS=1

"""

import datetime
import analysis_engine.get_task_results
import analysis_engine.work_tasks.custom_task
import analysis_engine.build_result as build_result
import analysis_engine.work_tasks.publish_pricing_update as \
    publisher
import analysis_engine.options_dates
import analysis_engine.get_pricing
from celery.task import task
from spylunking.log.setup_logging import build_colorized_logger
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import NOT_RUN
from analysis_engine.consts import ERR
from analysis_engine.consts import TICKER
from analysis_engine.consts import EMPTY_DF_STR
from analysis_engine.consts import get_status
from analysis_engine.consts import is_celery_disabled
from analysis_engine.consts import ppj
from analysis_engine.consts import ev

log = build_colorized_logger(
    name=__name__)


@task(
    bind=True,
    base=analysis_engine.work_tasks.custom_task.CustomTask,
    queue='handle_pricing_update_task')
def handle_pricing_update_task(
        self,
        work_dict):
    """handle_pricing_update_task

    Writes pricing updates to S3 and Redis

    :param work_dict: dictionary for key/values
    """

    label = 'update_prices'

    log.info(
        'task - {} - start'.format(
            label))

    ticker = TICKER
    ticker_id = 1
    rec = {
        'ticker': None,
        'ticker_id': None,
        'pricing_s3_bucket': None,
        'pricing_s3_key': None,
        'pricing_size': None,
        'pricing_redis_key': None,
        'news_s3_bucket': None,
        'news_s3_key': None,
        'news_size': None,
        'news_redis_key': None,
        'options_s3_bucket': None,
        'options_s3_key': None,
        'options_size': None,
        'options_redis_key': None
    }
    res = build_result.build_result(
        status=NOT_RUN,
        err=None,
        rec=rec)

    try:
        ticker = work_dict.get(
            'ticker',
            TICKER)
        ticker_id = int(work_dict.get(
            'ticker_id',
            1))

        rec['ticker'] = ticker
        rec['ticker_id'] = ticker_id

        pricing_data = work_dict['pricing']
        news_data = work_dict['news']
        options_data = work_dict['options']
        calls_data = options_data.get(
            'calls',
            EMPTY_DF_STR)
        puts_data = options_data.get(
            'puts',
            EMPTY_DF_STR)
        updated = work_dict['updated']
        label = work_dict.get(
            'label',
            label)

        cur_date = datetime.datetime.utcnow()
        cur_date_str = cur_date.strftime(
            '%Y_%m_%d_%H_%M_%S')

        pricing_s3_key = work_dict.get(
            'pricing_s3_key',
            'pricing_ticker_{}_id_{}_date_{}'.format(
                ticker,
                ticker_id,
                cur_date_str))
        news_s3_key = work_dict.get(
            'news_s3_key',
            'news_ticker_{}_id_{}_date_{}'.format(
                ticker,
                ticker_id,
                cur_date_str))
        options_s3_key = work_dict.get(
            'options_s3_key',
            'options_ticker_{}_id_{}_date_{}'.format(
                ticker,
                ticker_id,
                cur_date_str))
        calls_s3_key = work_dict.get(
            'calls_s3_key',
            'calls_ticker_{}_id_{}_date_{}'.format(
                ticker,
                ticker_id,
                cur_date_str))
        puts_s3_key = work_dict.get(
            'puts_s3_key',
            'puts_ticker_{}_id_{}_date_{}'.format(
                ticker,
                ticker_id,
                cur_date_str))

        pricing_s3_bucket = work_dict.get(
            'pricing_s3_bucket',
            'pricing')
        news_s3_bucket = work_dict.get(
            'news_s3_bucket',
            'news')
        options_s3_bucket = work_dict.get(
            'options_s3_bucket',
            'options')

        pricing_by_ticker_redis_key = work_dict.get(
            'pricing_redis_key',
            'price_{}'.format(
                ticker))
        news_by_ticker_redis_key = work_dict.get(
            'news_redis_key',
            'news_{}'.format(
                ticker))
        options_by_ticker_redis_key = work_dict.get(
            'options_redis_key',
            'options_{}'.format(
                ticker))
        calls_by_ticker_redis_key = work_dict.get(
            'calls_redis_key',
            'calls_{}'.format(
                ticker))
        puts_by_ticker_redis_key = work_dict.get(
            'puts_redis_key',
            'puts_{}'.format(
                ticker))

        pricing_size = len(str(
            pricing_data))
        news_size = len(str(
            news_data))
        options_size = len(str(
            options_data))
        calls_size = len(str(
            calls_data))
        puts_size = len(str(
            puts_data))

        payloads_to_publish = [
            {
                'ticker': ticker,
                'ticker_id': ticker_id,
                's3_bucket': pricing_s3_bucket,
                's3_key': pricing_s3_key,
                'data': pricing_data,
                'redis_key': pricing_by_ticker_redis_key,
                'size': pricing_size,
                'updated': updated,
                'label': label
            },
            {
                'ticker': ticker,
                'ticker_id': ticker_id,
                's3_bucket': options_s3_bucket,
                's3_key': options_s3_key,
                'data': options_data,
                'redis_key': options_by_ticker_redis_key,
                'size': options_size,
                'updated': updated,
                'label': label
            },
            {
                'ticker': ticker,
                'ticker_id': ticker_id,
                's3_bucket': options_s3_bucket,
                's3_key': calls_s3_key,
                'data': calls_data,
                'redis_key': calls_by_ticker_redis_key,
                'size': calls_size,
                'updated': updated,
                'label': label
            },
            {
                'ticker': ticker,
                'ticker_id': ticker_id,
                's3_bucket': options_s3_bucket,
                's3_key': puts_s3_key,
                'data': puts_data,
                'redis_key': puts_by_ticker_redis_key,
                'size': puts_size,
                'updated': updated,
                'label': label
            },
            {
                'ticker': ticker,
                'ticker_id': ticker_id,
                's3_bucket': news_s3_bucket,
                's3_key': news_s3_key,
                'data': news_data,
                'redis_key': news_by_ticker_redis_key,
                'size': news_size,
                'updated': updated,
                'label': label
            }
        ]

        rec['pricing_s3_bucket'] = pricing_s3_bucket
        rec['pricing_s3_key'] = pricing_s3_key
        rec['pricing_redis_key'] = pricing_by_ticker_redis_key
        rec['news_s3_bucket'] = news_s3_bucket
        rec['news_s3_key'] = news_s3_key
        rec['news_redis_key'] = news_by_ticker_redis_key
        rec['options_s3_bucket'] = options_s3_bucket
        rec['options_s3_key'] = options_s3_bucket
        rec['options_redis_key'] = options_by_ticker_redis_key

        total_payloads = len(payloads_to_publish)

        log.info(
            '{} ticker={} processing payloads={}'.format(
                label,
                ticker,
                total_payloads))

        for ridx, r in enumerate(payloads_to_publish):
            log.info(
                '{} ticker={} update={}/{} key={} redis_key={}'.format(
                    label,
                    ticker,
                    ridx,
                    total_payloads,
                    r['s3_key'],
                    r['redis_key']))
            r['celery_disabled'] = False
            r['label'] = 'handle_pricing_update_task-{}'.format(
                label)
            payload_res = \
                publisher.task_publish_pricing_update(
                    work_dict=r)
            log.info(
                '{} ticker={} update={}/{} status={} '
                's3_key={} redis_key={}'.format(
                    label,
                    ticker,
                    ridx,
                    total_payloads,
                    get_status(status=payload_res['status']),
                    r['s3_key'],
                    r['redis_key']))
        # end of for all payloads to publish

        res = build_result.build_result(
            status=SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        res = build_result.build_result(
            status=ERR,
            err=(
                'failed - handle_pricing_update_task '
                'dict={} with ex={}').format(
                    work_dict,
                    e),
            rec=rec)
        log.error(
            '{} - {}'.format(
                label,
                res['err']))
    # end of try/ex

    log.info(
        'task - handle_pricing_update_task done - '
        '{} - status={}'.format(
            label,
            get_status(res['status'])))

    return analysis_engine.get_task_results.get_task_results(
        work_dict=work_dict,
        result=res)
# end of handle_pricing_update_task


def run_handle_pricing_update_task(
        work_dict):
    """run_handle_pricing_update_task

    Celery wrapper for running without celery

    :param work_dict: task data
    """

    label = work_dict.get(
        'label',
        '')

    log.info(
        'run_handle_pricing_update_task - {} - start'.format(
            label))

    response = build_result.build_result(
        status=NOT_RUN,
        err=None,
        rec={})
    task_res = {}

    log.info(
        'run_handle_pricing_update_task - {} - done '
        'status={} err={} rec={}'.format(
            label,
            get_status(response['status']),
            response['err'],
            response['rec']))

    # allow running without celery
    if is_celery_disabled(
            work_dict=work_dict):
        work_dict['celery_disabled'] = True
        task_res = handle_pricing_update_task(
            work_dict)
        if task_res:
            response = task_res.get(
                'result',
                task_res)
            if ev('DEBUG_RESULTS', '0') == '1':
                response_details = response
                try:
                    response_details = ppj(response)
                except Exception:
                    response_details = response
                log.info(
                    '{} handle_pricing_update_task '
                    'task result={}'.format(
                        label,
                        response_details))
        else:
            log.error(
                '{} celery was disabled but the task={} '
                'did not return anything'.format(
                    label,
                    response))
        # end of if response
    else:
        task_res = handle_pricing_update_task.delay(
            work_dict=work_dict)
        rec = {
            'task_id': task_res
        }
        response = build_result.build_result(
            status=SUCCESS,
            err=None,
            rec=rec)
    # if celery enabled

    if response:
        if ev('DEBUG_RESULTS', '0') == '1':
            log.info(
                'run_handle_pricing_update_task - {} - done '
                'status={} err={} rec={}'.format(
                    label,
                    get_status(response['status']),
                    response['err'],
                    response['rec']))
        else:
            log.info(
                'run_handle_pricing_update_task - {} - done '
                'status={} err={}'.format(
                    label,
                    get_status(response['status']),
                    response['err']))
    else:
        log.info(
            'run_handle_pricing_update_task - {} - done '
            'no response'.format(
                label))
    # end of if/else response

    return response
# end of run_handle_pricing_update_task



"""
**Get New Pricing Data Task**

This will fetch data (pricing, financials, earnings, dividends, options,
and more) from these sources:

#.  Yahoo

#.  IEX

**Detailed example for getting new pricing data**

::

    import datetime
    from analysis_engine.api_requests
    import build_get_new_pricing_request
    from analysis_engine.work_tasks.get_new_pricing_data
    import get_new_pricing_data

    # store data
    cur_date = datetime.datetime.now().strftime('%Y-%m-%d')
    work = build_get_new_pricing_request(
        label='get-pricing-{}'.format(cur_date))
    work['ticker'] = 'TSLA'
    work['s3_bucket'] = 'pricing'
    work['s3_key'] = '{}-{}'.format(
        work['ticker'],
        cur_date)
    work['redis_key'] = '{}-{}'.format(
        work['ticker'],
        cur_date)
    work['celery_disabled'] = True
    res = get_new_pricing_data(
        work)
    print('full result dictionary:')
    print(res)
    if res['data']:
        print(
            'named datasets returned as '
            'json-serialized pandas DataFrames:')
        for k in res['data']:
            print(' - {}'.format(k))

.. warning:: When fetching pricing data from sources like IEX,
             Please ensure the returned values are
             not serialized pandas Dataframes to prevent
             issues with celery task results. Instead
             it is preferred to returned a ``df.to_json()``
             before sending the results into the
             results backend.

.. tip:: This task uses the `analysis_engine.work_tasks.
    custom_task.CustomTask class <https://github.com/A
    lgoTraders/stock-analysis-engine/blob/master/anal
    ysis_engine/work_tasks/custom_task.py>`__ for
    task event handling.

**Sample work_dict request for this method**

`analysis_engine.api_requests.build_get_new_pricing_request <https://
github.com/AlgoTraders/stock-analysis-engine/blob/master/ana
lysis_engine/api_requests.py#L49>`__

**Supported Environment Variables**

::

    export DEBUG_RESULTS=1

"""

import datetime
import copy
import analysis_engine.build_result as build_result
import analysis_engine.get_task_results
import analysis_engine.work_tasks.custom_task
import analysis_engine.options_dates
import analysis_engine.work_tasks.publish_pricing_update as \
    publisher
import analysis_engine.yahoo.get_data as yahoo_data
import analysis_engine.iex.get_data as iex_data
import analysis_engine.send_to_slack as slack_utils
from celery.task import task
from spylunking.log.setup_logging import build_colorized_logger
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import NOT_RUN
from analysis_engine.consts import ERR
from analysis_engine.consts import NOT_SET
from analysis_engine.consts import TICKER
from analysis_engine.consts import TICKER_ID
from analysis_engine.consts import EMPTY_DF_STR
from analysis_engine.consts import ENABLED_S3_UPLOAD
from analysis_engine.consts import S3_ACCESS_KEY
from analysis_engine.consts import S3_SECRET_KEY
from analysis_engine.consts import S3_REGION_NAME
from analysis_engine.consts import S3_ADDRESS
from analysis_engine.consts import S3_SECURE
from analysis_engine.consts import S3_BUCKET
from analysis_engine.consts import S3_KEY
from analysis_engine.consts import ENABLED_REDIS_PUBLISH
from analysis_engine.consts import REDIS_ADDRESS
from analysis_engine.consts import REDIS_KEY
from analysis_engine.consts import REDIS_PASSWORD
from analysis_engine.consts import REDIS_DB
from analysis_engine.consts import REDIS_EXPIRE
from analysis_engine.consts import get_status
from analysis_engine.consts import ev
from analysis_engine.consts import ppj
from analysis_engine.consts import is_celery_disabled
from analysis_engine.consts import FETCH_MODE_ALL
from analysis_engine.consts import FETCH_MODE_YHO
from analysis_engine.consts import FETCH_MODE_IEX
from analysis_engine.consts import DATASET_COLLECTION_VERSION
from analysis_engine.utils import utc_now_str
from analysis_engine.iex.consts import DEFAULT_FETCH_DATASETS
from analysis_engine.iex.consts import get_ft_str


log = build_colorized_logger(
    name=__name__)


@task(
    bind=True,
    base=analysis_engine.work_tasks.custom_task.CustomTask,
    queue='get_new_pricing_data')
def get_new_pricing_data(
        self,
        work_dict):
    """get_new_pricing_data

    Get Ticker information on:

    - prices - turn off with ``work_dict.get_pricing = False``
    - news - turn off with ``work_dict.get_news = False``
    - options - turn off with ``work_dict.get_options = False``

    :param work_dict: dictionary for key/values
    """

    label = 'get_new_pricing_data'

    log.info(
        'task - {} - start '
        'work_dict={}'.format(
            label,
            work_dict))

    ticker = TICKER
    ticker_id = TICKER_ID
    rec = {
        'pricing': None,
        'options': None,
        'calls': None,
        'puts': None,
        'news': None,
        'daily': None,
        'minute': None,
        'quote': None,
        'stats': None,
        'peers': None,
        'iex_news': None,
        'financials': None,
        'earnings': None,
        'dividends': None,
        'company': None,
        'exp_date': None,
        'publish_pricing_update': None,
        'date': utc_now_str(),
        'updated': None,
        'version': DATASET_COLLECTION_VERSION
    }
    res = {
        'status': NOT_RUN,
        'err': None,
        'rec': rec
    }

    try:
        ticker = work_dict.get(
            'ticker',
            ticker)
        ticker_id = work_dict.get(
            'ticker_id',
            TICKER_ID)
        s3_bucket = work_dict.get(
            's3_bucket',
            S3_BUCKET)
        s3_key = work_dict.get(
            's3_key',
            S3_KEY)
        redis_key = work_dict.get(
            'redis_key',
            REDIS_KEY)
        exp_date = work_dict.get(
            'exp_date',
            None)
        cur_date = datetime.datetime.utcnow()
        cur_strike = work_dict.get(
            'strike',
            None)
        contract_type = str(work_dict.get(
            'contract',
            'C')).upper()
        label = work_dict.get(
            'label',
            label)
        iex_datasets = work_dict.get(
            'iex_datasets',
            DEFAULT_FETCH_DATASETS)
        fetch_mode = work_dict.get(
            'fetch_mode',
            FETCH_MODE_ALL)

        # control flags to deal with feed issues:
        get_yahoo_data = True
        get_iex_data = True

        if (
                fetch_mode == FETCH_MODE_ALL
                or str(fetch_mode).lower() == 'all'):
            get_yahoo_data = True
            get_iex_data = True
        elif (
                fetch_mode == FETCH_MODE_YHO
                or str(fetch_mode).lower() == 'yahoo'):
            get_yahoo_data = True
            get_iex_data = False
        elif (
                fetch_mode == FETCH_MODE_IEX
                or str(fetch_mode).lower() == 'iex'):
            get_yahoo_data = False
            get_iex_data = True
        else:
            log.debug(
                '{} - unsupported fetch_mode={} value'.format(
                    label,
                    fetch_mode))

        if not exp_date:
            exp_date = analysis_engine.options_dates.option_expiration(
                date=exp_date)
        else:
            exp_date = datetime.datetime.strptime(
                exp_date,
                '%Y-%m-%d')

        rec['updated'] = cur_date.strftime('%Y-%m-%d %H:%M:%S')
        log.info(
            '{} getting pricing for ticker={} '
            'cur_date={} exp_date={} '
            'yahoo={} iex={}'.format(
                label,
                ticker,
                cur_date,
                exp_date,
                get_yahoo_data,
                get_iex_data))

        yahoo_rec = {
            'ticker': ticker,
            'pricing': None,
            'options': None,
            'calls': None,
            'puts': None,
            'news': None,
            'exp_date': None,
            'publish_pricing_update': None,
            'date': None,
            'updated': None
        }

        if get_yahoo_data:
            log.info(
                '{} yahoo ticker={}'.format(
                    label,
                    ticker))
            yahoo_res = yahoo_data.get_data_from_yahoo(
                work_dict=work_dict)
            if yahoo_res['status'] == SUCCESS:
                yahoo_rec = yahoo_res['rec']
                log.info(
                    '{} yahoo ticker={} '
                    'status={} err={}'.format(
                        label,
                        ticker,
                        get_status(status=yahoo_res['status']),
                        yahoo_res['err']))
                rec['pricing'] = yahoo_rec.get('pricing', '{}')
                rec['news'] = yahoo_rec.get('news', '{}')
                rec['options'] = yahoo_rec.get('options', '{}')
                rec['calls'] = rec['options'].get('calls', EMPTY_DF_STR)
                rec['puts'] = rec['options'].get('puts', EMPTY_DF_STR)
            else:
                log.error(
                    '{} failed YAHOO ticker={} '
                    'status={} err={}'.format(
                        label,
                        ticker,
                        get_status(status=yahoo_res['status']),
                        yahoo_res['err']))
        # end of get from yahoo

        if get_iex_data:
            num_iex_ds = len(iex_datasets)
            log.debug(
                '{} iex datasaets={}'.format(
                    label,
                    num_iex_ds))
            for idx, ft_type in enumerate(iex_datasets):
                dataset_field = get_ft_str(ft_type=ft_type)

                log.info(
                    '{} iex={}/{} field={} ticker={}'.format(
                        label,
                        idx,
                        num_iex_ds,
                        dataset_field,
                        ticker))
                iex_label = '{}-{}'.format(
                    label,
                    dataset_field)
                iex_req = copy.deepcopy(work_dict)
                iex_req['label'] = iex_label
                iex_req['ft_type'] = ft_type
                iex_req['field'] = dataset_field
                iex_req['ticker'] = ticker
                iex_res = iex_data.get_data_from_iex(
                    work_dict=iex_req)

                if iex_res['status'] == SUCCESS:
                    iex_rec = iex_res['rec']
                    log.info(
                        '{} iex ticker={} field={} '
                        'status={} err={}'.format(
                            label,
                            ticker,
                            dataset_field,
                            get_status(status=iex_res['status']),
                            iex_res['err']))
                    if dataset_field == 'news':
                        rec['iex_news'] = iex_rec['data']
                    else:
                        rec[dataset_field] = iex_rec['data']
                else:
                    log.debug(
                        '{} failed IEX ticker={} field={} '
                        'status={} err={}'.format(
                            label,
                            ticker,
                            dataset_field,
                            get_status(status=iex_res['status']),
                            iex_res['err']))
                # end of if/else succcess
            # end idx, ft_type in enumerate(iex_datasets):
        # end of if get_iex_data

        update_req = {
            'data': rec
        }
        update_req['ticker'] = ticker
        update_req['ticker_id'] = ticker_id
        update_req['strike'] = cur_strike
        update_req['contract'] = contract_type
        update_req['s3_enabled'] = work_dict.get(
            's3_enabled',
            ENABLED_S3_UPLOAD)
        update_req['redis_enabled'] = work_dict.get(
            'redis_enabled',
            ENABLED_REDIS_PUBLISH)
        update_req['s3_bucket'] = s3_bucket
        update_req['s3_key'] = s3_key
        update_req['s3_access_key'] = work_dict.get(
            's3_access_key',
            S3_ACCESS_KEY)
        update_req['s3_secret_key'] = work_dict.get(
            's3_secret_key',
            S3_SECRET_KEY)
        update_req['s3_region_name'] = work_dict.get(
            's3_region_name',
            S3_REGION_NAME)
        update_req['s3_address'] = work_dict.get(
            's3_address',
            S3_ADDRESS)
        update_req['s3_secure'] = work_dict.get(
            's3_secure',
            S3_SECURE)
        update_req['redis_key'] = redis_key
        update_req['redis_address'] = work_dict.get(
            'redis_address',
            REDIS_ADDRESS)
        update_req['redis_password'] = work_dict.get(
            'redis_password',
            REDIS_PASSWORD)
        update_req['redis_db'] = int(work_dict.get(
            'redis_db',
            REDIS_DB))
        update_req['redis_expire'] = work_dict.get(
            'redis_expire',
            REDIS_EXPIRE)
        update_req['updated'] = rec['updated']
        update_req['label'] = label
        update_req['celery_disabled'] = True
        update_status = NOT_SET

        try:
            update_res = publisher.run_publish_pricing_update(
                work_dict=update_req)
            update_status = update_res.get(
                'status',
                NOT_SET)
            if ev('DEBUG_RESULTS', '0') == '1':
                log.info(
                    '{} update_res status={} data={}'.format(
                        label,
                        get_status(status=update_status),
                        ppj(update_res)))
            else:
                log.info(
                    '{} run_publish_pricing_update status={}'.format(
                        label,
                        get_status(status=update_status)))
            # end of if/else

            rec['publish_pricing_update'] = update_res
            res = build_result.build_result(
                status=SUCCESS,
                err=None,
                rec=rec)
        except Exception as f:
            err = (
                '{} publisher.run_publish_pricing_update failed '
                'with ex={}'.format(
                    label,
                    f))
            log.error(err)
            res = build_result.build_result(
                status=ERR,
                err=err,
                rec=rec)
        # end of trying to publish results to connected services

    except Exception as e:
        res = build_result.build_result(
            status=ERR,
            err=(
                'failed - get_new_pricing_data '
                'dict={} with ex={}').format(
                    work_dict,
                    e),
            rec=rec)
        log.error(
            '{} - {}'.format(
                label,
                res['err']))
    # end of try/ex

    if ev('DATASET_COLLECTION_SLACK_ALERTS', '0') == '1':
        env_name = 'DEV'
        if ev('PROD_SLACK_ALERTS', '1') == '1':
            env_name = 'PROD'
        done_msg = (
            'Dataset collected ticker=*{}* on env=*{}* '
            'redis_key={} s3_key={} iex={} yahoo={}'.format(
                ticker,
                env_name,
                redis_key,
                s3_key,
                get_iex_data,
                get_yahoo_data))
        log.debug(
            '{} sending slack msg={}'.format(
                label,
                done_msg))
        if res['status'] == SUCCESS:
            slack_utils.post_success(
                msg=done_msg,
                block=False,
                jupyter=True)
        else:
            slack_utils.post_failure(
                msg=done_msg,
                block=False,
                jupyter=True)
        # end of if/else success
    # end of publishing to slack

    log.info(
        'task - get_new_pricing_data done - '
        '{} - status={}'.format(
            label,
            get_status(res['status'])))

    return analysis_engine.get_task_results.get_task_results(
        work_dict=work_dict,
        result=res)
# end of get_new_pricing_data


def run_get_new_pricing_data(
        work_dict):
    """run_get_new_pricing_data

    Celery wrapper for running without celery

    :param work_dict: task data
    """

    label = work_dict.get(
        'label',
        '')

    log.info(
        'run_get_new_pricing_data - {} - start'.format(
            label))

    response = build_result.build_result(
        status=NOT_RUN,
        err=None,
        rec={})
    task_res = {}

    # allow running without celery
    if is_celery_disabled(
            work_dict=work_dict):
        work_dict['celery_disabled'] = True
        task_res = get_new_pricing_data(
            work_dict)
        if task_res:
            response = task_res.get(
                'result',
                task_res)
            if ev('DEBUG_RESULTS', '0') == '1':
                response_details = response
                try:
                    response_details = ppj(response)
                except Exception:
                    response_details = response

                log.info(
                    '{} task result={}'.format(
                        label,
                        response_details))
        else:
            log.error(
                '{} celery was disabled but the task={} '
                'did not return anything'.format(
                    label,
                    response))
        # end of if response
    else:
        task_res = get_new_pricing_data.delay(
            work_dict=work_dict)
        rec = {
            'task_id': task_res
        }
        response = build_result.build_result(
            status=SUCCESS,
            err=None,
            rec=rec)
    # if celery enabled

    if response:
        if ev('DEBUG_RESULTS', '0') == '1':
            log.info(
                'run_get_new_pricing_data - {} - done '
                'status={} err={} rec={}'.format(
                    label,
                    get_status(response['status']),
                    response['err'],
                    response['rec']))
        else:
            log.info(
                'run_get_new_pricing_data - {} - done '
                'status={} err={}'.format(
                    label,
                    get_status(response['status']),
                    response['err']))
    else:
        log.info(
            'run_get_new_pricing_data - {} - done '
            'no response'.format(
                label))
    # end of if/else response

    return response
# end of run_get_new_pricing_data



"""
Charting functions with matplotlib, numpy, pandas, and seaborn

Change the footnote with:

::

    export PLOT_FOOTNOTE="custom footnote on images"

.. note: most of these functions were ported from
         the repo: https://github.com/jay-johnson/scipype

"""

import datetime
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import seaborn as sns
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def add_footnote(
        fig=None,
        xpos=0.90,
        ypos=0.01,
        text=None,
        color='#888888',
        fontsize=8):
    """add_footnote

    Add a footnote based off the environment key:
    ``PLOT_FOOTNOTE``

    :param fig: add the footnote to this figure object
    :param xpos: x-axes position
    :param ypos: y-axis position
    :param text: text in the footnote
    :param color: font color
    :param fontsize: text size for the footnote text
    """
    if not fig:
        return

    use_footnote = text
    if not use_footnote:
        use_footnote = ae_consts.ev(
            'PLOT_FOOTNOTE',
            'algotraders')

    fig.text(
        xpos,
        ypos,
        use_footnote,
        va='bottom',
        fontsize=fontsize,
        color=color)
# end of add_footnote


def set_common_seaborn_fonts():
    """set_common_seaborn_fonts

    Set the font and text style
    """
    sns.set(font='serif')
    sns.set_context(
        'paper',
        rc={
            'font.size': 12,
            'axes.titlesize': 12,
            'axes.labelsize': 10
        }
    )
# end of set_common_seaborn_fonts


def send_final_log(
        log_label,
        fn_name,
        result):
    """send_final_log

    :param log_label: log identifier
    :param fn_name: function name
    :param result: dictionary result
    """

    if not fn_name:
        log.error(
            '{} missing fn_name parameter for send_final_log'.format(
                log_label))
        return

    if not result:
        log.error(
            '{} missing result parameter for send_final_log'.format(
                log_label))
        return

    str_result = (
        '{} - {} - done status={} '
        'err={}'.format(
            log_label,
            fn_name,
            ae_consts.get_status(result['status']),
            result['err']))

    if result['status'] == ae_consts.ERR:
        log.error(str_result)
    else:
        log.info(str_result)
    # handle log based off status

# end of send_final_log


def show_with_entities(
        log_label,
        xlabel,
        ylabel,
        title,
        ax,
        fig,
        legend_list=None,
        show_plot=True):
    """show_with_entities

    Helper for showing a plot with a legend and a
    footnoe

    :param log_label: log identifier
    :param xlabel: x-axis label
    :param ylabel: y-axis label
    :param title: title of the plot
    :param ax: axes
    :param fig: figure
    :param legend_list: list of legend items to show
    :param show_plot: bool to show the plot
    """

    log.debug(
        '{} - '
        'show_with_entities'
        ' - start'.format(
            log_label))

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    ax.set_title(title)

    if not legend_list:
        ax.legend(
            loc="best",
            prop={
                "size": "medium"
            }
        )
    else:
        ax.legend(
            legend_list,
            loc="best",
            prop={
                "size": "medium"
            }
        )
    # end of if/else legend entries

    add_footnote(fig)
    plt.tight_layout()

    if show_plot:
        plt.show()
    else:
        plt.plot()
# end of show_with_entities


def plot_overlay_pricing_and_volume(
        log_label,
        ticker,
        df,
        xlabel=None,
        ylabel=None,
        high_color=ae_consts.PLOT_COLORS['high'],
        close_color=ae_consts.PLOT_COLORS['blue'],
        volume_color=ae_consts.PLOT_COLORS['green'],
        date_format=ae_consts.IEX_MINUTE_DATE_FORMAT,
        show_plot=True,
        dropna_for_all=True):
    """plot_overlay_pricing_and_volume

    Plot pricing (high, low, open, close) and volume as
    an overlay off the x-axis

    Here is a sample chart from the
    `Stock Analysis Jupyter Intro Notebook <https://github.com/Al
    goTraders/stock-analysis-engine/blob/master/co
    mpose/docker/notebooks/Stock-Analysis-Intro.ipynb>`__

    .. image:: https://i.imgur.com/pH368gy.png

    :param log_label: log identifier
    :param ticker: ticker name
    :param df: timeseries ``pandas.DateFrame``
    :param xlabel: x-axis label
    :param ylabel: y-axis label
    :param high_color: optional - high plot color
    :param close_color: optional - close plot color
    :param volume_color: optional - volume color
    :param data_format: optional - date format string this must
        be a valid value for the ``df['date']`` column
        that would work with:
        ``datetime.datetime.stftime(date_format)``
    :param show_plot: optional - bool to show the plot
    :param dropna_for_all: optional - bool to toggle keep None's in
        the plot ``df`` (default is drop them
        for display purposes)
    """

    rec = {
        'fig': None,
        'ax': None,
        'ax2': None
    }
    result = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    try:

        log.info(
            '{} - '
            'plot_overlay_pricing_and_volume'
            ' - start'.format(
                log_label))

        set_common_seaborn_fonts()

        use_df = df
        if dropna_for_all:
            log.info(
                '{} - '
                'plot_overlay_pricing_and_volume'
                ' - dropna_for_all'.format(
                    log_label))
            use_df = df.dropna(axis=0, how='any')
        # end of pre-plot dataframe scrubbing

        fig, ax = plt.subplots(
            sharex=True,
            sharey=True,
            figsize=(15.0, 10.0))
        ax.plot(
            use_df['date'],
            use_df['close'],
            color=close_color)
        ax.plot(
            use_df['date'],
            use_df['high'],
            color=high_color)

        # use a second axis to display the
        # volume since it is in a different range
        # this will fill under the
        # volume's y values as well
        ax2 = ax.twinx()
        ax2.plot(
            use_df['date'],
            use_df['volume'],
            linestyle='-',
            color=volume_color,
            alpha=0.6)
        ax2.fill_between(
            use_df['date'].values,
            0,
            use_df['volume'].values,
            color=volume_color,
            alpha=0.5)
        # setup the second plot for volume
        ax2.set_ylim([0, ax2.get_ylim()[1] * 3])

        plt.grid(True)
        use_xlabel = xlabel
        use_ylabel = ylabel
        if not use_xlabel:
            xlabel = 'Minute Dates'
        if not use_ylabel:
            ylabel = '{} High and Close Prices'.format(
                ticker)
        plt.xlabel(use_xlabel)
        plt.ylabel(use_ylabel)

        # Build a date vs Close DataFrame
        start_date = ''
        end_date = ''
        try:
            start_date = str(use_df.iloc[0]['date'].strftime(date_format))
            end_date = str(use_df.iloc[-1]['date'].strftime(date_format))
        except Exception as f:
            date_format = '%Y-%m-%d'
            start_date = str(use_df.iloc[0]['date'].strftime(date_format))
            end_date = str(use_df.iloc[-1]['date'].strftime(date_format))

        use_title = (
            '{} Pricing from: {} to {}'.format(
                ticker,
                start_date,
                end_date))
        ax.set_title(use_title)

        # Merge in the second axis (volume) Legend
        handles, labels = plt.gca().get_legend_handles_labels()
        newLabels, newHandles = [], []
        for handle, label in zip(handles, labels):
            if label not in newLabels:
                newLabels.append(label)
                newHandles.append(handle)
        lines = ax.get_lines() + ax2.get_lines() + newHandles
        ax.legend(
            lines,
            [l.get_label() for l in lines],
            loc='best',
            shadow=True)

        # Build out the xtick chart by the dates
        ax.xaxis.grid(True, which='minor')
        ax.fmt_xdata = mdates.DateFormatter(date_format)
        ax.xaxis.set_major_formatter(ax.fmt_xdata)
        ax.xaxis.set_minor_formatter(ax.fmt_xdata)

        # turn off the grids on volume
        ax2.fmt_xdata = mdates.DateFormatter(date_format)
        ax2.xaxis.grid(False)
        ax2.yaxis.grid(False)
        ax2.yaxis.set_ticklabels([])

        fig.autofmt_xdate()

        show_with_entities(
            log_label=log_label,
            xlabel=xlabel,
            ylabel=ylabel,
            title=use_title,
            ax=ax,
            fig=fig,
            show_plot=show_plot)

        rec['fig'] = fig
        rec['ax'] = ax
        rec['ax2'] = ax2

        result = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        err = (
            'failed plot_overlay_pricing_and_volume '
            'and volume with ex={}'.format(
                e))
        result = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=rec)
    # end of try/ex

    send_final_log(
        log_label=log_label,
        fn_name='plot_overlay_pricing_and_volume',
        result=result)

    return result
# end of plot_overlay_pricing_and_volume


def plot_hloc_pricing(
        log_label,
        ticker,
        df,
        title,
        show_plot=True,
        dropna_for_all=True):
    """plot_hloc_pricing

    Plot the high, low, open and close columns together on a chart

    :param log_label: log identifier
    :param ticker: ticker
    :param df: initialized ``pandas.DataFrame``
    :param title: title for the chart
    :param show_plot: bool to show the plot
    :param dropna_for_all: optional - bool to toggle keep None's in
                           the plot ``df`` (default is drop them
                           for display purposes)
    """

    rec = {
        'ax': None,
        'fig': None
    }
    result = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    try:

        log.info(
            '{} - '
            'plot_hloc_pricing'
            ' - start'.format(
                log_label))

        set_common_seaborn_fonts()

        fig, ax = plt.subplots(
            figsize=(15.0, 10.0))

        use_df = df
        if dropna_for_all:
            log.info(
                '{} - '
                'plot_hloc_pricing'
                ' - dropna_for_all'.format(
                    log_label))
            use_df = df.dropna(axis=0, how='any')
        # end of pre-plot dataframe scrubbing

        plt.plot(
            use_df['date'],
            use_df['high'],
            label='High',
            color=ae_consts.PLOT_COLORS['high'],
            alpha=0.4)
        plt.plot(
            use_df['date'],
            use_df['low'],
            label='Low',
            color=ae_consts.PLOT_COLORS['low'],
            alpha=0.4)
        plt.plot(
            use_df['date'],
            use_df['close'],
            label='Close',
            color=ae_consts.PLOT_COLORS['close'],
            alpha=0.4)
        plt.plot(
            use_df['date'],
            use_df['open'],
            label='Open',
            color=ae_consts.PLOT_COLORS['open'],
            alpha=0.4)

        xlabel = 'Dates'
        ylabel = 'Prices'

        plt.grid(True)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        # Build a date vs Close DataFrame
        start_date = str(use_df.iloc[0]['date'].strftime('%Y-%m-%d'))
        end_date = str(use_df.iloc[-1]['date'].strftime('%Y-%m-%d'))
        if not title:
            title = (
                '{} Pricing from: {} to {}'.format(
                    ticker,
                    start_date,
                    end_date))
        ax.set_title(title)

        # Build out the xtick chart by the dates
        ax.xaxis.grid(True, which='minor')

        legend_list = [
            'high',
            'low',
            'close',
            'open'
        ]

        fig.autofmt_xdate()
        show_with_entities(
            log_label=log_label,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            ax=ax,
            fig=fig,
            legend_list=legend_list,
            show_plot=show_plot)

        rec['ax'] = ax
        rec['fig'] = fig

        result = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        err = (
            'failed plot_hloc_pricing with ex={}'.format(
                e))
        log.error(err)
        result = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=rec)
    # end of try/ex

    send_final_log(
        log_label=log_label,
        fn_name='plot_hloc_pricing',
        result=result)

    return result
# end of plot_hloc_pricing


def plot_df(
        log_label,
        title,
        column_list,
        df,
        xcol='date',
        xlabel='Date',
        ylabel='Pricing',
        linestyle='-',
        color='blue',
        show_plot=True,
        dropna_for_all=True):
    """plot_df
    :param log_label: log identifier
    :param title: title of the plot
    :param column_list: list of columns in the df to show
    :param df: initialized ``pandas.DataFrame``
    :param xcol: x-axis column in the initialized ``pandas.DataFrame``
    :param xlabel: x-axis label
    :param ylabel: y-axis label
    :param linestyle: style of the plot line
    :param color: color to use
    :param show_plot: bool to show the plot
    :param dropna_for_all: optional - bool to toggle keep None's in
                           the plot ``df`` (default is drop them
                           for display purposes)
    """

    rec = {
        'ax': None,
        'fig': None
    }
    result = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    try:

        log.info(
            '{} - '
            'plot_df'
            ' - start'.format(
                log_label))

        use_df = df
        if dropna_for_all:
            log.info(
                '{} - '
                'plot_df'
                ' - dropna_for_all'.format(
                    log_label))
            use_df = df.dropna(axis=0, how='any')
        # end of pre-plot dataframe scrubbing

        set_common_seaborn_fonts()

        hex_color = ae_consts.PLOT_COLORS['blue']
        fig, ax = plt.subplots(figsize=(15.0, 10.0))

        if linestyle == '-':
            use_df[column_list].plot(
                x=xcol,
                linestyle=linestyle,
                ax=ax,
                color=hex_color,
                rot=0)
        else:
            use_df[column_list].plot(
                kind='bar',
                x=xcol,
                ax=ax,
                color=hex_color,
                rot=0)

        if 'date' in str(xcol).lower():
            # plot the column
            min_date = use_df.iloc[0][xcol]
            max_date = use_df.iloc[-1][xcol]

            # give a bit of space at each end of the plot - aesthetics
            span = max_date - min_date
            extra = int(span.days * 0.03) * datetime.timedelta(days=1)
            ax.set_xlim([min_date - extra, max_date + extra])

            # format the x tick marks
            ax.xaxis.set_minor_formatter(mdates.DateFormatter('%b\n%Y'))
            ax.xaxis.set_major_locator(plt.NullLocator())
            ax.xaxis.set_minor_locator(
                mdates.MonthLocator(
                    bymonthday=1,
                    interval=1))

            # grid, legend and yLabel
            ax.xaxis.grid(True, which='minor')

        # end of date formatting

        show_with_entities(
            log_label=log_label,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            ax=ax,
            fig=fig,
            show_plot=show_plot)

        rec['ax'] = ax
        rec['fig'] = fig

        result = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        err = (
            'failed plot_df title={} with ex={}'.format(
                title,
                e))
        result = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=rec)
    # end of try/ex

    send_final_log(
        log_label=log_label,
        fn_name='plot_df',
        result=result)

    return result
# end of plot_df


def dist_plot(
        log_label,
        df,
        width=10.0,
        height=10.0,
        title='Distribution Plot',
        style='default',
        xlabel='',
        ylabel='',
        show_plot=True,
        dropna_for_all=True):
    """dist_plot

    Show a distribution plot for the passed in dataframe: ``df``

    :param log_label: log identifier
    :param df: initialized ``pandas.DataFrame``
    :param width: width of figure
    :param height: height of figure
    :param style: style to use
    :param xlabel: x-axis label
    :param ylabel: y-axis label
    :param show_plot: bool to show plot or not
    :param dropna_for_all: optional - bool to toggle keep None's in
                           the plot ``df`` (default is drop them
                           for display purposes)
    """

    rec = {}
    result = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    try:

        log.info(
            '{} - '
            'dist_plot'
            ' - start'.format(
                log_label))

        set_common_seaborn_fonts()

        fig, ax = plt.subplots(
            figsize=(width, height))

        if style == 'default':
            sns.set_context('poster')
            sns.axes_style('darkgrid')

        use_df = df
        if dropna_for_all:
            log.info(
                '{} - '
                'dist_plot'
                ' - dropna_for_all'.format(
                    log_label))
            use_df = df.dropna(axis=0, how='any')
        # end of pre-plot dataframe scrubbing

        sns.distplot(
            use_df,
            color=ae_consts.PLOT_COLORS['blue'],
            ax=ax)

        if xlabel != '':
            ax.set_xlabel(xlabel)
        if ylabel != '':
            ax.set_ylabel(ylabel)

        plt.tight_layout()
        plt.subplots_adjust(top=0.9)
        fig.suptitle(title)

        show_with_entities(
            log_label=log_label,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            ax=ax,
            fig=fig,
            show_plot=show_plot)

        rec['ax'] = ax
        rec['fig'] = fig

        result = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        err = (
            'failed dist_plot title={} with ex={}'.format(
                title,
                e))
        log.error(err)
        result = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=rec)
    # end of try/ex

    send_final_log(
        log_label=log_label,
        fn_name='dist_plot',
        result=result)

    return result
# end of dist_plot



"""
Helper for setting data in redis

Debug redis calls with:

::

    export DEBUG_REDIS=1

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json

"""
import json
import redis
import analysis_engine.build_result as build_result
from spylunking.log.setup_logging import build_colorized_logger
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import NOT_RUN
from analysis_engine.consts import ERR
from analysis_engine.consts import ev

log = build_colorized_logger(
    name=__name__)


def set_data_in_redis_key(
        label=None,
        data=None,
        client=None,
        host=None,
        port=None,
        password=None,
        db=None,
        key=None,
        expire=None,
        px=None,
        nx=False,
        xx=False,
        serializer='json',
        encoding='utf-8'):
    """set_data_in_redis_key

    :param label: log tracking label
    :param data: data to set in redis
    :param client: initialized redis client
    :param host: not used yet - redis host
    :param port: not used yet - redis port
    :param password: not used yet - redis password
    :param db: not used yet - redis db
    :param key: not used yet - redis key
    :param expire: redis expire
    :param px: redis px
    :param nx: redis nx
    :param xx: redis xx
    :param serializer: not used yet - support for future
                       pickle objects in redis
    :param encoding: format of the encoded key in redis
    """

    data_str = None
    encoded_data = None

    rec = {}
    res = build_result.build_result(
        status=NOT_RUN,
        err=None,
        rec=rec)

    log_id = label if label else 'set-redis'

    try:
        log.debug(
            '{} serializer={} encoding={} for key={}'.format(
                log_id,
                serializer,
                encoding,
                key))
        if serializer == 'json':
            data_str = json.dumps(data)
            encoded_data = data_str.encode(encoding)
        else:
            encoded_data = None
            err = (
                '{} unsupported serializer={} '
                'encoding={} key={}'.format(
                    log_id,
                    serializer,
                    encoding,
                    key))
            log.error(err)
            res = build_result.build_result(
                status=ERR,
                err=err,
                rec=rec)
            return res
        # if supported serializer

        if encoded_data:
            if ev('DEBUG_REDIS', '0') == '1':
                log.debug(
                    '{} set - key={} data={}'.format(
                        log_id,
                        key,
                        encoded_data))

            use_client = client
            if not use_client:
                log.debug(
                    '{} set key={} new client={}:{}@{}'.format(
                        log_id,
                        key,
                        host,
                        port,
                        db))
                use_client = redis.Redis(
                    host=host,
                    port=port,
                    password=password,
                    db=db)
            else:
                log.debug(
                    '{} set key={} client'.format(
                        log_id,
                        key))
            # create Redis client if not set

            use_client.set(
                name=key,
                value=encoded_data,
                ex=expire,
                px=px,
                nx=nx,
                xx=xx)
            res = build_result.build_result(
                status=SUCCESS,
                err=None,
                rec=rec)
            return res
        else:
            err = (
                '{} no data for key={}'.format(
                    log_id,
                    key))
            log.error(err)
            res = build_result.build_result(
                status=ERR,
                err=err,
                rec=rec)
            return res
        # end of if have data to set
    except Exception as e:
        err = (
            '{} failed - redis set from data={} encoded_data={} '
            'key={} ex={}'.format(
                log_id,
                data,
                encoded_data,
                key,
                e))
        log.error(err)
        res = build_result.build_result(
            status=ERR,
            err=err,
            rec=rec)
    # end of try/ex for setting redis data

    return res
# end of set_data_in_redis_key



"""
Plot a ``Trading History`` dataset using seaborn and matplotlib
"""

import datetime
import matplotlib.pyplot as plt
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.charts as ae_charts
import analysis_engine.build_result as build_result
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def plot_trading_history(
        title,
        df,
        red,
        red_color=None,
        red_label=None,
        blue=None,
        blue_color=None,
        blue_label=None,
        green=None,
        green_color=None,
        green_label=None,
        orange=None,
        orange_color=None,
        orange_label=None,
        date_col='date',
        xlabel='Date',
        ylabel='Algo Values',
        linestyle='-',
        width=15.0,
        height=15.0,
        date_format='%d\n%b',
        df_filter=None,
        start_date=None,
        footnote_text=None,
        footnote_xpos=0.70,
        footnote_ypos=0.01,
        footnote_color='#888888',
        footnote_fontsize=8,
        scale_y=False,
        show_plot=True,
        dropna_for_all=False,
        verbose=False):
    """plot_trading_history

    Plot columns up to 4 lines from the ``Trading History`` dataset

    :param title: title of the plot
    :param df: dataset which is ``pandas.DataFrame``
    :param red: string - column name to plot in
        ``red_color`` (or default ``ae_consts.PLOT_COLORS[red]``)
        where the column is in the ``df`` and
        accessible with:``df[red]``
    :param red_color: hex color code to plot the data in the
        ``df[red]``  (default is ``ae_consts.PLOT_COLORS['red']``)
    :param red_label: optional - string for the label used
        to identify the ``red`` line in the legend
    :param blue: string - column name to plot in
        ``blue_color`` (or default ``ae_consts.PLOT_COLORS['blue']``)
        where the column is in the ``df`` and
        accessible with:``df[blue]``
    :param blue_color: hex color code to plot the data in the
        ``df[blue]``  (default is ``ae_consts.PLOT_COLORS['blue']``)
    :param blue_label: optional - string for the label used
        to identify the ``blue`` line in the legend
    :param green: string - column name to plot in
        ``green_color`` (or default ``ae_consts.PLOT_COLORS['darkgreen']``)
        where the column is in the ``df`` and
        accessible with:``df[green]``
    :param green_color: hex color code to plot the data in the
        ``df[green]``  (default is ``ae_consts.PLOT_COLORS['darkgreen']``)
    :param green_label: optional - string for the label used
        to identify the ``green`` line in the legend
    :param orange: string - column name to plot in
        ``orange_color`` (or default ``ae_consts.PLOT_COLORS['orange']``)
        where the column is in the ``df`` and
        accessible with:``df[orange]``
    :param orange_color: hex color code to plot the data in the
        ``df[orange]``  (default is ``ae_consts.PLOT_COLORS['orange']``)
    :param orange_label: optional - string for the label used
        to identify the ``orange`` line in the legend
    :param date_col: string - date column name
        (default is ``date``)
    :param xlabel: x-axis label
    :param ylabel: y-axis label
    :param linestyle: style of the plot line
    :param width: float - width of the image
    :param height: float - height of the image
    :param date_format: string - format for dates
    :param df_filter: optional - initialized ``pandas.DataFrame`` query
        for reducing the ``df`` records before plotting. As an eaxmple
        ``df_filter=(df['close'] > 0.01)`` would find only records in
        the ``df`` with a ``close`` value greater than ``0.01``
    :param start_date: optional - string ``datetime``
        for plotting only from a date formatted as
        ``YYYY-MM-DD HH\:MM\:SS``
    :param footnote_text: optional - string footnote text
        (default is ``algotraders <DATE>``)
    :param footnote_xpos: optional - float for footnote position
        on the x-axies
        (default is ``0.75``)
    :param footnote_ypos: optional - float for footnote position
        on the y-axies
        (default is ``0.01``)
    :param footnote_color: optional - string hex color code for
        the footnote text
        (default is ``#888888``)
    :param footnote_fontsize: optional - float footnote font size
        (default is ``8``)
    :param scale_y: optional - bool to scale the y-axis with
        .. code-block:: python

            use_ax.set_ylim(
                [0, use_ax.get_ylim()[1] * 3])
    :param show_plot: bool to show the plot
    :param dropna_for_all: optional - bool to toggle keep None's in
        the plot ``df`` (default is drop them for display purposes)
    :param verbose: optional - bool to show logs for debugging
        a dataset
    """

    rec = {
        'ax1': None,
        'ax2': None,
        'ax3': None,
        'ax4': None,
        'fig': None
    }
    result = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    if verbose:
        log.info('plot_trading_history - start')

    use_red = red_color
    use_blue = blue_color
    use_green = green_color
    use_orange = orange_color

    if not use_red:
        use_red = ae_consts.PLOT_COLORS['red']
    if not use_blue:
        use_blue = ae_consts.PLOT_COLORS['blue']
    if not use_green:
        use_green = ae_consts.PLOT_COLORS['darkgreen']
    if not use_orange:
        use_orange = ae_consts.PLOT_COLORS['orange']

    use_footnote = footnote_text
    if not use_footnote:
        use_footnote = (
            'algotraders - {}'.format(
                datetime.datetime.now().strftime(
                    ae_consts.COMMON_TICK_DATE_FORMAT)))

    column_list = [
        date_col
    ]
    all_plots = []
    if red:
        column_list.append(red)
        all_plots.append({
            'column': red,
            'color': use_red})
    if blue:
        column_list.append(blue)
        all_plots.append({
            'column': blue,
            'color': use_blue})
    if green:
        column_list.append(green)
        all_plots.append({
            'column': green,
            'color': use_green})
    if orange:
        column_list.append(orange)
        all_plots.append({
            'column': orange,
            'color': use_orange})

    use_df = df
    if start_date:
        start_date_value = datetime.datetime.strptime(
            start_date,
            ae_consts.COMMON_TICK_DATE_FORMAT)
        use_df = df[(df[date_col] >= start_date_value)][column_list]
    # end of filtering by start date

    if verbose:
        log.info(
            'plot_history_df start_date={} df.index={} column_list={}'.format(
                start_date,
                len(use_df.index),
                column_list))

    if hasattr(df_filter, 'to_json'):
        # Was seeing this warning below in Jupyter:
        # UserWarning: Boolean Series key
        # will be reindexed to match DataFrame index
        # use_df = use_df[df_filter][column_list]
        # now using:
        use_df = use_df.loc[df_filter, column_list]

    if verbose:
        log.info(
            'plot_history_df filter df.index={} column_list={}'.format(
                start_date,
                len(use_df.index),
                column_list))

    if dropna_for_all:
        use_df = use_df.dropna(axis=0, how='any')
        if verbose:
            log.info('plot_history_df dropna_for_all')
    # end of pre-plot dataframe scrubbing

    ae_charts.set_common_seaborn_fonts()

    hex_color = ae_consts.PLOT_COLORS['blue']
    fig, ax = plt.subplots(
        sharex=True,
        sharey=True,
        figsize=(
            width,
            height))

    # Convert matplotlib date numbers to strings for dates to
    # avoid dealing with weekend date gaps in plots
    date_strings, date_labels = \
        ae_utils.get_trade_open_xticks_from_date_col(
            use_df[date_col])

    """
    hit the slice warning with this approach before
    and one trying df[date_col] = df[date_col].dt.strftime

    SettingWithCopyWarning
    Try using .loc[row_indexer,col_indexer] = value instead

    use_df[date_col].replace(
        use_df[date_col].dt.strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT),
        inplace=True)
    trying this:
    https://stackoverflow.com/questions/19738169/
    convert-column-of-date-objects-in-pandas-dataframe-to-strings
    """
    use_df[date_col] = use_df[date_col].apply(lambda x: x.strftime(
        ae_consts.COMMON_TICK_DATE_FORMAT))

    all_axes = []
    num_plots = len(all_plots)
    for idx, node in enumerate(all_plots):
        column_name = node['column']
        hex_color = node['color']

        use_ax = ax
        if idx > 0:
            use_ax = ax.twinx()

        if verbose:
            log.info(
                'plot_history_df - {}/{} - {} in {} - ax={}'.format(
                    (idx + 1),
                    num_plots,
                    column_name,
                    hex_color,
                    use_ax))

        all_axes.append(use_ax)
        use_ax.plot(
            use_df[date_col],
            use_df[column_name],
            linestyle=linestyle,
            color=hex_color)
        if idx > 0:
            if scale_y:
                use_ax.set_ylim(
                    [0, use_ax.get_ylim()[1] * 3])
            use_ax.yaxis.set_ticklabels([])
            use_ax.yaxis.set_ticks([])
            use_ax.xaxis.grid(False)
            use_ax.yaxis.grid(False)
        # end if this is not the fist axis

        use_ax.set_xticks(date_strings)
        use_ax.set_xticklabels(date_labels, rotation=45, ha='right')
    # end of for all plots

    lines = []
    for idx, cur_ax in enumerate(all_axes):
        ax_lines = cur_ax.get_lines()
        for line in ax_lines:
            label_name = str(line.get_label())
            use_label = label_name
            if idx == 0:
                if red_label:
                    use_label = red_label
            elif idx == 1:
                if blue_label:
                    use_label = blue_label
            elif idx == 2:
                use_label = label_name[-20:]
                if green_label:
                    use_label = green_label
            elif idx == 3:
                use_label = label_name[-20:]
                if orange_label:
                    use_label = orange_label
            else:
                if len(label_name) > 10:
                    use_label = label_name[-20:]
            # end of fixing the labels in the legend
            line.set_label(use_label)
            if line.get_label() not in lines:
                lines.append(line)
        rec['ax{}'.format(idx + 1)] = cur_ax
    # end of compiling a new-shortened legend while removing dupes

    for idx, cur_ax in enumerate(all_axes):
        if cur_ax:
            if cur_ax.get_legend():
                cur_ax.get_legend().remove()
    # end of removing all previous legends

    if verbose:
        log.info(
            'legend lines={}'.format(
                [l.get_label() for l in lines]))
    # log what's going to be in the legend

    ax.legend(
        lines,
        [l.get_label() for l in lines],
        loc='best',
        shadow=True)

    fig.autofmt_xdate()

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    ax.set_title(title)
    ae_charts.add_footnote(
        fig=fig,
        xpos=footnote_xpos,
        ypos=footnote_ypos,
        text=use_footnote,
        color=footnote_color,
        fontsize=footnote_fontsize)
    plt.tight_layout()

    plt.show()
    if show_plot:
        plt.show()
    else:
        plt.plot()

    rec['fig'] = fig

    result = build_result.build_result(
        status=ae_consts.SUCCESS,
        err=None,
        rec=rec)

    return result
# end of plot_history_df
